# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/unet_qc.ipynb.

# %% auto 0
__all__ = ['UNet_block', 'Encoder', 'Decoder', 'QC_Cond_UNet_config', 'QC_Cond_UNet', 'QC_Compilation_UNet_config',
           'QC_Compilation_UNet']

# %% ../../src/models/unet_qc.ipynb 3
from ..imports import *
from .config_model import Config_Model
import genQC.models.layers as layers
import genQC.models.transformers as transformers
from .unitary_encoder import Unitary_encoder, Unitary_encoder_config

# %% ../../src/models/unet_qc.ipynb 5
class UNet_block(nn.Module):
    """The basic block of the U-Net. Is conditioned via cross-attention in `SpatialTransformer` and addition of the time ebedding in `ResBlock2D_Conditional`."""
    def __init__(self, ch_in, ch_out, t_emb_size, cond_emb_size, num_heads=8, num_res_blocks=1, transformer_depth=1):
        super().__init__()
                                
        self.resBlocks = nn.ModuleList() 
        for i in range(num_res_blocks):          
            self.resBlocks.append(layers.ResBlock2D_Conditional(ch_in, ch_out, t_emb_size, kernel_size=(1, 3)))
            ch_in = ch_out 
            
        self.transformer_depth = transformer_depth
        if self.transformer_depth > 0:
            self.spatialTransformer = transformers.SpatialTransformer(ch_out, cond_emb_size, num_heads, transformer_depth)
                   
        self._init_weights()
          
    def _init_weights(self):
        for resBlock in self.resBlocks:
            resBlock.conv2.weight.data.zero_() 
                      
    def forward(self, x, t_emb, c_emb, attn_mask, key_padding_mask):
        for resBlock in self.resBlocks:
            x = resBlock(x, t_emb)
        
        if self.transformer_depth > 0:
            x = self.spatialTransformer(x, c_emb, attn_mask, key_padding_mask)   
            
        return x

# %% ../../src/models/unet_qc.ipynb 6
class Encoder(nn.Module):
    """Encoder definition of the U-Net."""
    
    def __init__(self, model_features, t_emb_size, cond_emb_size, num_heads, num_res_blocks, transformer_depths):
        super().__init__()
        self.enc_blocks = nn.ModuleList() 
        
        in_ch = model_features[0]
        for model_feature,heads,res_blocks,transformer_depth in zip(model_features[1:-1], num_heads[:-1], num_res_blocks[:-1], transformer_depths[:-1]):
            out_ch         = model_feature            
            enc_block      = UNet_block(in_ch, out_ch, t_emb_size, cond_emb_size, heads, res_blocks, transformer_depth)            
            enc_block.down = layers.DownBlock2D(out_ch, out_ch, kernel_size=(1, 2), stride=(1, 2), padding=(0,0))       
            in_ch          = out_ch 
        
            self.enc_blocks.append(enc_block)
    
        self.mid_block = UNet_block(in_ch, model_features[-1], t_emb_size, cond_emb_size, num_heads[-1], num_res_blocks[-1], transformer_depths[-1])  # should be!
        
    def forward(self, x, t_emb, c_emb, attn_mask=None, key_padding_mask=None):        
        ftrs = []       
        for i,enc_block in enumerate(self.enc_blocks):
            x = enc_block(x, t_emb, c_emb, attn_mask[i], key_padding_mask[i])  # attn_mask only in first layer!

            ftrs.append(x)  
            x = enc_block.down(x)
    
        x = self.mid_block(x, t_emb, c_emb, attn_mask[-1], key_padding_mask[-1])
        ftrs.append(x)
        
        return ftrs

# %% ../../src/models/unet_qc.ipynb 7
class Decoder(nn.Module):
    """Decoder definition of the U-Net."""
    
    def __init__(self, model_features, t_emb_size, cond_emb_size, num_heads, num_res_blocks, transformer_depths):
        super().__init__()
        self.dec_blocks = nn.ModuleList()
                     
        in_ch = model_features[0]
        for model_feature,heads,res_blocks, transformer_depth in zip(model_features[1:], num_heads[1:], num_res_blocks[1:], transformer_depths[1:]):
            out_ch       = model_feature           
            dec_block    = UNet_block(out_ch*2, out_ch, t_emb_size, cond_emb_size, heads, res_blocks, transformer_depth)            
            dec_block.up = layers.UpBlock2D(in_ch, out_ch, kernel_size=(1, 2), stride=(1, 2), padding=(0,0))                          
            in_ch        = out_ch 
            
            self.dec_blocks.append(dec_block) 
                          
    def forward(self, x, encoder_features, t_emb, c_emb, attn_mask=None, key_padding_mask=None): 
        for i,(dec_block, ftr) in enumerate(zip(self.dec_blocks, encoder_features)):         
            x = dec_block.up(x)
            x = torch.cat([x, ftr / (2.0**0.5)], dim=1)
            
            x = dec_block(x, t_emb, c_emb, attn_mask[i], key_padding_mask[i])                                
        return x

# %% ../../src/models/unet_qc.ipynb 9
@dataclass
class QC_Cond_UNet_config:  
    model_features: list[int]
    clr_dim: int
    num_clrs: int
    t_emb_size: int  
    cond_emb_size: int
    num_heads: list[int]
    num_res_blocks: list[int]
    transformer_depths: list[int]

# %% ../../src/models/unet_qc.ipynb 10
class QC_Cond_UNet(Config_Model):
    """Conditional U-Net model for quantum circuits. Implemets `embedd_clrs` and `invert_clr` functions to embed and decode color-tensors."""
    
    def __init__(self, model_features=[32,32,64], clr_dim=8, num_clrs=8, t_emb_size=128, cond_emb_size=512, 
                 num_heads=[8,8,2], num_res_blocks=[2, 2, 4], transformer_depths=[1,2,1]):
        
        super().__init__()       
        
        self.clr_dim  = clr_dim     
        self.num_clrs = num_clrs
        
        self.t_emb_size    = model_features[0] * 4 if not t_emb_size else t_emb_size
        self.cond_emb_size = model_features[0] * 4 if not cond_emb_size else cond_emb_size
            
        self.params_config = QC_Cond_UNet_config(model_features, self.clr_dim, self.num_clrs, self.t_emb_size, self.cond_emb_size, num_heads, num_res_blocks, transformer_depths)
            
        #-----------    
                             
        self.enc_chs = [model_features[0]] + list(model_features)
        self.dec_chs = list(model_features)[::-1]
        
        #-----------
                
        self.t_emb   = layers.TimeEmbedding(d_model=self.t_emb_size)
        self.emb_clr = nn.Embedding(num_embeddings=self.num_clrs, embedding_dim=self.clr_dim)           
        
        self.conv_in = nn.Conv2d(self.clr_dim, model_features[0], kernel_size=1, stride=1, padding ="same") #was kernel_size=3
        self.pos_enc = layers.PositionalEncoding2D(d_model=model_features[0])

        self.encoder = Encoder(self.enc_chs, self.t_emb_size, cond_emb_size=self.cond_emb_size, num_heads=num_heads, num_res_blocks=num_res_blocks, transformer_depths=transformer_depths)
        self.decoder = Decoder(self.dec_chs, self.t_emb_size, cond_emb_size=self.cond_emb_size, num_heads=num_heads[::-1], num_res_blocks=num_res_blocks[::-1], transformer_depths=transformer_depths[::-1])
        self.head    = nn.Conv2d(self.dec_chs[-1], self.clr_dim, kernel_size=1, stride=1, padding ="same")
                                               
        self._init_weights()
          
    def _init_weights(self):
        self.emb_clr.weight.requires_grad = False
        nn.init.orthogonal_(self.emb_clr.weight.data)
        
        for enc_block in self.encoder.enc_blocks:
            nn.init.orthogonal_(enc_block.down.conv1.weight)
        
        for dec_block in self.decoder.dec_blocks:
            nn.init.orthogonal_(dec_block.up.conv1.weight)
        
        self.head.weight.data.zero_()
       
    #--------------------------------------------
    
    def embedd_clrs(self, x):
        sign = torch.sign(x + 0.1)  #trick: add 0.1 so that the sign of 0 is +1, else the 0 token would be all 0s.     
        clr  = self.emb_clr(torch.abs(x))      
        x = clr * sign[:, :, :, None]        
        x = torch.permute(x, (0, 3, 1, 2))       
        return x
    
    @torch.no_grad()
    def invert_clr(self, x):
        #collaps clr to gate ... use cos sim
        
        clrs = self.emb_clr.weight.detach() # is [clr_num, clr_dim]
        
        model_device = clrs.device
        input_device = x.device
        
        # to shape [b*space*time, clr_dim]
        x      = x.to(model_device)
        x_flat = x.permute(0, 2, 3, 1).reshape(-1, x.shape[1])
                         
        #normlize for cos sim       
        norm_clr    = clrs   / torch.linalg.vector_norm(  clrs, dim=1, keepdim=True) 
        norm_x_flat = x_flat / torch.linalg.vector_norm(x_flat, dim=1, keepdim=True) 
        
        #matmul out is [clr_num, b*space*time]
        sim = torch.matmul(norm_clr, norm_x_flat.T) 
            
        #get highest abs(similarity) and sign of it
        abs_sim = sim.abs()
        max_idx = torch.argmax(abs_sim, dim=0) #reduce the clr_num dim
        sign = torch.sign(sim[max_idx, torch.arange(x_flat.shape[0])])
        scores_flat = max_idx * sign

        # back to [b, space, time]
        scores = scores_flat.reshape(x.shape[0], x.shape[2], x.shape[3]).to(torch.int64)      
        scores = scores.to(input_device)
        
        return scores
    
    #--------------------------------------------
    
    def forward(self, x, t, c_emb, attn_mask=None, key_padding_mask=None):
        if attn_mask        is None: attn_mask        = [None] * len(self.enc_chs)
        if key_padding_mask is None: key_padding_mask = [None] * len(self.enc_chs)
              
        t_emb = self.t_emb(t)
                
        x = self.conv_in(x)            
        x = self.pos_enc(x) 
        
        enc_ftrs = self.encoder(x, t_emb=t_emb, c_emb=c_emb, attn_mask=attn_mask, key_padding_mask=key_padding_mask)[::-1]
        out      = self.decoder(x=enc_ftrs[0], encoder_features=enc_ftrs[1:], t_emb=t_emb, c_emb=c_emb, 
                                attn_mask=attn_mask[::-1][1:], key_padding_mask=key_padding_mask[::-1][1:])
        out      = self.head(out)       
        return out

# %% ../../src/models/unet_qc.ipynb 12
@dataclass
class QC_Compilation_UNet_config(QC_Cond_UNet_config):  
    unitary_encoder_config: Unitary_encoder_config

# %% ../../src/models/unet_qc.ipynb 13
class QC_Compilation_UNet(QC_Cond_UNet):
    """Extension of the `QC_Cond_UNet` to accept unitary conditions."""
    
    def __init__(self, model_features=[32,32,64], clr_dim=8, num_clrs=8, t_emb_size=128, cond_emb_size=512, 
                 num_heads=[8,8,2], num_res_blocks=[2, 2, 4], transformer_depths=[1,2,1], unitary_encoder_config=None): 
        
        super().__init__(model_features, clr_dim, num_clrs, t_emb_size, cond_emb_size, num_heads, num_res_blocks, transformer_depths)

        if is_dataclass(unitary_encoder_config):
            unitary_encoder_config = asdict(unitary_encoder_config)
        self.unitary_encoder = Unitary_encoder(**unitary_encoder_config)
        self.params_config   = QC_Compilation_UNet_config(model_features, self.clr_dim, self.num_clrs, self.t_emb_size, self.cond_emb_size, num_heads, num_res_blocks, transformer_depths, self.unitary_encoder.params_config)
    
    def forward(self, x, t, c_emb, U, attn_mask=None, key_padding_mask=None):
        u_emb = self.unitary_encoder(U)            # [batch, seq2, ch]     
        c_emb = torch.cat([c_emb, u_emb], dim=1)   # [batch, seq1+seq2, ch]  
        out = super().forward(x, t, c_emb, attn_mask, key_padding_mask)
        return out



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/util.ipynb.

# %% auto 0
__all__ = ['MemoryCleaner', 'virtual', 'DataLoaders', 'infer_torch_device', 'number_of_paramters', 'normalize_tensor',
           'scale_tensor', 'savePdf', 'savePng', 'saveSvg', 'plot_image_grid', 'latents_to_pil']

# %% ../src/util.ipynb 3
from .imports import *
import gc, sys, traceback

# %% ../src/util.ipynb 5
class MemoryCleaner():
    """CLass with static methods to clean (gpu) memory."""
    
    @staticmethod
    def _clean_ipython_hist():
        # Code in this function mainly copied from IPython source
        if not 'get_ipython' in globals(): return
        ip = get_ipython()
        user_ns = ip.user_ns
        ip.displayhook.flush()
        pc = ip.displayhook.prompt_count + 1
        for n in range(1, pc): user_ns.pop('_i'+repr(n),None)
        user_ns.update(dict(_i='',_ii='',_iii=''))
        hm = ip.history_manager
        hm.input_hist_parsed[:] = [''] * pc
        hm.input_hist_raw[:] = [''] * pc
        hm._i = hm._ii = hm._iii = hm._i00 =  ''

    @staticmethod
    def _clean_tb():
        if hasattr(sys, 'last_traceback'):
            traceback.clear_frames(sys.last_traceback)
            delattr(sys, 'last_traceback')
        if hasattr(sys, 'last_type'): delattr(sys, 'last_type')
        if hasattr(sys, 'last_value'): delattr(sys, 'last_value')
        
    @staticmethod
    def purge_mem():
        """Clear all. Purge all memory."""
        MemoryCleaner._clean_tb()
        MemoryCleaner._clean_ipython_hist()
        gc.collect()
        torch.cuda.empty_cache()    

# %% ../src/util.ipynb 7
def virtual(f: callable) -> callable:
    '''Decorator to enfore subclass method implementations and raises error at method calls.'''
    @functools.wraps(f)
    def inner(self, *args, **kwargs): raise NotImplementedError(f"Virtual method {f.__name__} needs to be implemented by subclass {self.__class__.__name__}.")      
    return inner

# %% ../src/util.ipynb 10
class DataLoaders:
    """Combines train and valid `DataLoader`."""
    def __init__(self, *dls: list[DataLoader]): self.train, self.valid = dls[:2]

# %% ../src/util.ipynb 11
def infer_torch_device(): 
    if torch.cuda.is_available():   
        torch.backends.cudnn.benchmark = True

        dev_cap = torch.cuda.get_device_capability()
        
        if dev_cap[0] >= 8: # AMPERE
            print(f"[INFO]: Cuda device has a capability of {dev_cap[0]}.{dev_cap[1]} (>= 8), allowing tf32 matmul.")
            torch.backends.cuda.matmul.allow_tf32 = True
        
        return torch.device("cuda")
    return torch.device("cpu")

# %% ../src/util.ipynb 13
def number_of_paramters(model: nn.Module): return sum([p.flatten().shape[0] for p in model.parameters()])

# %% ../src/util.ipynb 14
def normalize_tensor(t: torch.Tensor):
    """[0,1] to [-1,1]"""
    return t * 2.0 - 1.0

def scale_tensor(t: torch.Tensor):
    """[-1,1] to [0,1]"""
    return (t / 2.0 + 0.5).clamp(0.0, 1.0)

# %% ../src/util.ipynb 16
def savePdf(filename): plt.savefig(filename + '.pdf', bbox_inches='tight')
def savePng(filename): plt.savefig(filename + '.png', bbox_inches='tight')
def saveSvg(filename): plt.savefig(filename + '.svg', bbox_inches='tight')

# %% ../src/util.ipynb 17
def plot_image_grid(imgs: Union[list, np.array, torch.Tensor], labels: list=None, labels_fs="medium", 
                    figsize=(16, 4), cols=8, cmap="Greys", show_colorbar=False, **imshow_kwargs):  
    if type(imgs) is list: n = len(imgs)
    elif type(imgs) is np.ndarray: n = imgs.shape[0]
    elif type(imgs) is torch.Tensor: n = imgs.shape[0]
    else: raise NotImplementedError("err type:", type(imgs))

    if n == 0: return

    cols = min(n, cols)
    rows = math.ceil(n/cols)

    fig, axs = plt.subplots(rows, cols, figsize=figsize, squeeze=False, constrained_layout=True)   
    for i, (r, c) in enumerate(itertools.product(range(rows), range(cols))):
        plt.sca(axs[r,c])
        plt.axis('off')
         
        if i >= n: continue

        if labels is not None: plt.title(labels[i], fontsize=labels_fs)
        p = plt.imshow(imgs[i], cmap=cmap, **imshow_kwargs) #cmap ignored for RGB   
        if show_colorbar: plt.colorbar(p)
        
    plt.show()

# %% ../src/util.ipynb 19
def latents_to_pil(latents:torch.Tensor, channels=None):
    if channels is None:        
        channels = latents.shape[1] if len(latents.shape) > 3 else 1
            
    images = scale_tensor(latents)
    images = images.detach().cpu().permute(0, 2, 3, 1).numpy()
    
    if channels == 1: images = images[:, :, :, 0]

    images = (images * 255).round().astype(np.uint8)
  
    pil_images = [Image.fromarray(image) for image in images]
    return pil_images   



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/transformers.ipynb.

# %% auto 0
__all__ = ['BasisSelfAttnBlock', 'BasisCrossAttnBlock', 'SpatialTransformerSelfAttn', 'SpatialTransformer']

# %% ../../src/models/transformers.ipynb 2
from ..imports import *
import genQC.models.layers as layers

# %% ../../src/models/transformers.ipynb 4
class BasisSelfAttnBlock(nn.Module):
    """A self attention block, i.e. a `transformer` encoder."""
    def __init__(self, ch, num_heads, dropout=0):
        super().__init__()
        self.self_att  = nn.MultiheadAttention(ch, num_heads=num_heads, batch_first=False) #[t, b, c]
        self.ff    = layers.FeedForward(ch, ch)   
        self.norm1 = nn.LayerNorm(ch)
        self.norm2 = nn.LayerNorm(ch)
        self.drop  = nn.Dropout(dropout)
               
    def forward(self, x, attn_mask=None, key_padding_mask=None, need_weights=False):
        #x     ... [  t, batch, ch]       
        #c_emb ... [seq, batch, ch]
        
        self_out    = self.norm1(x)  
        self_out, _ = self.self_att(self_out, key=self_out, value=self_out, attn_mask=attn_mask, key_padding_mask=key_padding_mask, need_weights=need_weights)
        self_out    = self.drop(self_out) + x      
        
        feed_out = self.norm2(self_out)              
        feed_out = self.ff(feed_out)
        feed_out = self.drop(feed_out) + self_out            
                   
        return feed_out     

# %% ../../src/models/transformers.ipynb 5
class BasisCrossAttnBlock(nn.Module):
    """A cross attention block, i.e. a `transformer` decoder."""
    def __init__(self, ch, cond_emb_size, num_heads, dropout=0.0):
        super().__init__()
        self.self_att  = nn.MultiheadAttention(ch, num_heads=num_heads, batch_first=False) #[t, b, c]
        self.cross_att = nn.MultiheadAttention(ch, num_heads=num_heads, batch_first=False) 
        self.ff    = layers.FeedForward(ch, ch)   
        self.norm1 = nn.LayerNorm(ch)
        self.norm2 = nn.LayerNorm(ch)
        self.norm3 = nn.LayerNorm(ch)
        self.drop  = nn.Dropout(dropout)
        
    def forward(self, x, c_emb, attn_mask=None, key_padding_mask=None, need_weights=False):
        #x     ... [  t, batch, ch]       
        #c_emb ... [seq, batch, ch]
        
        self_out    = self.norm1(x)  
        self_out, _ = self.self_att(self_out, key=self_out, value=self_out, attn_mask=attn_mask, key_padding_mask=key_padding_mask, need_weights=need_weights)
        self_out    = self.drop(self_out) + x      
        
        cross_out    = self.norm2(self_out)   
        cross_out, _ = self.cross_att(cross_out, key=c_emb, value=c_emb, need_weights=need_weights)
        cross_out    = self.drop(cross_out) + self_out         
        
        feed_out = self.norm3(cross_out)              
        feed_out = self.ff(feed_out)
        feed_out = self.drop(feed_out) + cross_out            
                   
        return feed_out     

# %% ../../src/models/transformers.ipynb 7
class SpatialTransformerSelfAttn(nn.Module):
    """A spatial residual `transformer`, only uses self-attention."""
    def __init__(self, ch, num_heads, depth, dropout=0.0):
        super().__init__()       
        self.norm               = torch.nn.GroupNorm(num_groups=32, num_channels=ch, eps=1e-6, affine=True)
        self.transformer_blocks = nn.ModuleList([BasisSelfAttnBlock(ch, num_heads, dropout) for d in range(depth)])
        
    def forward(self, x, attn_mask=None, key_padding_mask=None):
        #x      ... [batch, ch, space, time]  
        #c_emb  ... [batch, seq, ch]
        b, ch, space, time = x.shape
            
        x_in = x
        
        #-------------------------
        x = self.norm(x) 
        
        x = torch.reshape(x, (b, ch, space*time))
        x = torch.permute(x, (2, 0, 1)).contiguous()           # to [t, batch, ch]    
        
        #-------------------------   
        # x = self.proj_in(x) #NEW   only used so that ch is a multiple of heads
        
        for transformer_block in self.transformer_blocks:
            x = transformer_block(x, attn_mask, key_padding_mask)
                
        # feed_out = self.proj_out(feed_out) #NEW
        #-------------------------
            
        x = torch.permute(x, (1, 2, 0))              # back to [batch, ch, t] 
        x = torch.reshape(x, (b, ch, space, time)).contiguous()
                
        return x + x_in

# %% ../../src/models/transformers.ipynb 8
class SpatialTransformer(nn.Module):
    """A spatial residual `transformer`, uses self- and cross-attention on conditional input."""
    
    def __init__(self, ch, cond_emb_size, num_heads, depth, dropout=0.0):
        super().__init__()       
        self.cat_proj           = nn.Linear(cond_emb_size, ch)  
        self.norm               = torch.nn.GroupNorm(num_groups=32, num_channels=ch, eps=1e-6, affine=True)
        self.transformer_blocks = nn.ModuleList([BasisCrossAttnBlock(ch, cond_emb_size, num_heads, dropout) for d in range(depth)])
        
    def forward(self, x, c_emb, attn_mask=None, key_padding_mask=None):
        #x      ... [batch, ch, space, time]  
        #c_emb  ... [batch, seq, ch]
        b, ch, space, time = x.shape
            
        x_in = x
        
        #-------------------------
        x = self.norm(x) 
        
        x = torch.reshape(x, (b, ch, space*time))
        x = torch.permute(x, (2, 0, 1)).contiguous()           # to [t, batch, ch]    
       
        c_emb = self.cat_proj(c_emb)        
        c_emb = torch.permute(c_emb, (1, 0, 2)).contiguous()  # to [seq, batch, ch]
        
        #-------------------------   
        # x = self.proj_in(x) #NEW   only used so that ch is a multiple of heads
        
        for transformer_block in self.transformer_blocks:
            x = transformer_block(x, c_emb, attn_mask, key_padding_mask)
                
        # feed_out = self.proj_out(feed_out) #NEW
        #-------------------------
            
        x = torch.permute(x, (1, 2, 0))              # back to [batch, ch, t] 
        x = torch.reshape(x, (b, ch, space, time)).contiguous()
                
        return x + x_in



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/platform/qcircuit_dataset_construction.ipynb.

# %% auto 0
__all__ = ['get_target_control_qubits', 'encode_circuit', 'decode_circuit', 'get_rnd_encoded_circuit', 'get_rnd_encoded_circuits',
           'gen_qc_dataset', 'get_specific_rnd_srv_circuit', 'gen_compilation_rndGates_dataset']

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 3
from ..imports import *
from .simulation.qcircuit_sim import *
import genQC.dataset.dataset_helper as dahe

import qiskit.quantum_info as qi
from qiskit import QuantumCircuit
from qiskit.circuit.gate import Gate
import qiskit.circuit.library as ql

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 5
def get_target_control_qubits(qc: QuantumCircuit, gate: Gate):
    acts_on_cnt = gate.operation.num_qubits
    acts_on     = [qc.find_bit(qubit).index for qubit in gate.qubits]      # order: (*control_qubits, *target_qubits)

    assert acts_on_cnt == len(acts_on), "error in: acts_on_cnt == len(acts_on)"

    num_ctrl_qubits = gate.operation.num_ctrl_qubits if hasattr(gate.operation, "num_ctrl_qubits") else 0 
    num_targ_qubits = acts_on_cnt - num_ctrl_qubits

    control_qubits, target_qubits = acts_on[:-num_targ_qubits], acts_on[-num_targ_qubits:]   
    
    return control_qubits, target_qubits  

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 6
def encode_circuit(qc: QuantumCircuit, num_of_qubits, gate_classes: dict, max_gates: int, sign_labels={"control_qubits":-1, "target_qubits":+1}, return_params=False):
    # circuit tensor
    #  [qbits, time]  .. in +- gate_number   
    #                         0 for empty
       
    tensor = torch.zeros((num_of_qubits, max_gates), dtype=torch.int32) 
    params = []
    
    for t, gate in enumerate(qc.data):
        params.append(gate.operation.params)
        
        gate_id = gate_classes[gate.operation.name]  #for new tensor just use this as the abs(T) and then assign the sign dep on the c/t
                         
        control_qubits, target_qubits = get_target_control_qubits(qc, gate)  
                       
        for bit in control_qubits:
            tensor[bit, t] = gate_id * sign_labels["control_qubits"]
        
        for bit in target_qubits:
            tensor[bit, t] = gate_id * sign_labels["target_qubits"]

    if return_params: 
        num_of_max_params = max(len(para) for para in params)
        params_tensor     = torch.zeros((num_of_max_params, max_gates), dtype=torch.float32)
        
        for t, para in enumerate(params):
            params_tensor[:len(para), t] = torch.tensor(para)
                     
        return tensor, params_tensor
        
    return tensor

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 7
def decode_circuit(enc_tensor: torch.Tensor, gate_pool: list[Gate], place_barrier=True, sign_labels={"control_qubits":-1, "target_qubits":+1}, params_tensor=None):  
    # should have dim 2, [bits, sequence]
    #minus ... control_qubits
    #plus  ... target_qubits

    assert enc_tensor.ndim == 2, f"{enc_tensor.shape=}"
    num_of_qubits, time = enc_tensor.shape
        
    gate_qiskit_classes = {(i+1):x for i,x in enumerate(gate_pool)}
    
    qc = QuantumCircuit(num_of_qubits)  
          
    for t in range(time):         
        enc_time_slice = enc_tensor[:, t] # only contains all bits at time t   
                      
        for gate_index,gate_qiskit_class in gate_qiskit_classes.items():                       
            target_qubits  = (enc_time_slice == (sign_labels["target_qubits"] *gate_index)).nonzero()
            control_qubits = (enc_time_slice == (sign_labels["control_qubits"]*gate_index)).nonzero()
                       
            if target_qubits.nelement() > 0:
                num_of_paramters = get_number_of_gate_params(gate_qiskit_class)   
                if exists(params_tensor) and num_of_paramters > 0 : params = params_tensor[:num_of_paramters, t].tolist()
                else:                                               params = [0] * num_of_paramters
                
                qc.append(gate_qiskit_class(*params), [*control_qubits.tolist(), *target_qubits.tolist()], [])        
                if place_barrier: qc.barrier()
                break #break on first hit, per def only one gate allowed per t
                
            elif control_qubits.nelement() > 0: #no target but control means error
                raise RuntimeError("control_qubits.nelement() > 0")
            #else we are fine with tensor that have time steps with no action!

    return qc

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 10
def get_rnd_encoded_circuit(num_of_qubits, min_gates, max_gates, gate_pool, gate_classes, rng, optimized=True, return_params=False):    
    qc = rnd_circuit(num_of_qubits, rng.integers(min_gates, max_gates+1), gate_pool, rng)  
    if optimized: qc = optimize_circuit(qc, gate_pool)        
    svr = schmidt_rank_vector(qi.DensityMatrix(qc))
    
    if return_params:
        qc_tensor, params_tensor = encode_circuit(qc, num_of_qubits, gate_classes, max_gates, return_params=return_params)  
        return qc, qc_tensor, svr, params_tensor
    
    qc_tensor = encode_circuit(qc, num_of_qubits, gate_classes, max_gates, return_params=return_params)     
    return qc, qc_tensor, svr 

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 11
def get_rnd_encoded_circuits(samples, num_of_qubits=3, min_gates=3, max_gates=10, gate_pool=[ql.HGate, ql.CXGate], optimized=True, silent=False, return_params=False):
    gate_classes = gate_pool_to_gate_classes(gate_pool) #{x().name:(i+1) for i,x in enumerate(gate_pool)}
    
    rng = np.random.default_rng()
    
    data   = []
    label  = []
    params = []
    
    for i in tqdm(range(samples), disable=silent):            
        if return_params:
            qc, qc_tensor, svr, params_tensor = get_rnd_encoded_circuit(num_of_qubits, min_gates, max_gates, gate_pool, gate_classes, rng, optimized, return_params=return_params)  
            params.append(params_tensor)
            
        else:
             qc, qc_tensor, svr = get_rnd_encoded_circuit(num_of_qubits, min_gates, max_gates, gate_pool, gate_classes, rng, optimized, return_params=return_params)  
            
        data.append(qc_tensor)
        label.append(svr)
    
    if return_params: return data, label, params
    return data, label

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 13
def gen_qc_dataset(samples, num_of_qubits, min_gates, max_gates, gate_pool, optimized, silent=False):
    tensor, srv = get_rnd_encoded_circuits(samples, num_of_qubits, min_gates, max_gates, gate_pool, optimized, silent)

    # make sure we have unique circuits
    tensor = torch.stack(tensor, dim=0)
    tensor_unique, tensor_indices = dahe.get_unique_elements_indices(tensor)
    
    if not silent: print(f"Generated unique circuits: {tensor_unique.shape[0]}")
    
    #--------------------------
    #select uniques only
    
    x = tensor[tensor_indices]
    y = torch.Tensor(srv).type(torch.int32)[tensor_indices] #leave as tensor, treat as 2D condition, combine cond into one large (each cat)
    
    return x,y

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 15
def get_specific_rnd_srv_circuit(srv, requested_length, gate_pool, max_i=2000, silent=True, fix_length_after_optimizing=True, requested_length_tolerance=0):
    rng = np.random.default_rng()
    
    num_of_qubits = len(srv) 
    is_srv        = None

    if requested_length < sum(srv)-num_of_qubits: return None   #not possible to generate this srv, to few gates
    
    i = 0
    while is_srv != srv:                  # brute-force sample a SRV
        qc = rnd_circuit(num_of_qubits, requested_length, gate_pool, rng)  
        qc = optimize_circuit(qc, gate_pool)        

        if i > max_i: 
            if not silent: print(f"Max i reached: {srv=}  {requested_length=}  {requested_length_tolerance=}  {max_i=}")
            return None #raise RuntimeError("max i reached")     
        i += 1

        #---------------
        if fix_length_after_optimizing and len(qc.data) < requested_length-requested_length_tolerance:
                continue
        
        is_srv = schmidt_rank_vector(qi.DensityMatrix(qc)) 
       
    return qc

# %% ../../src/platform/qcircuit_dataset_construction.ipynb 17
def gen_compilation_rndGates_dataset(samples, num_of_qubits, min_gates, max_gates, gate_pool, min_sub_gate_pool_cnt=1, silent=False):
    '''Samples rnd circuit with a rnd subset of gates and return qc with gate label and unitary'''

    gate_classes = {x().name:(i+1) for i,x in enumerate(gate_pool)} #+1 for empty! global gate classes so we fix the indices! 1...H   2...CX   so on

    #-------------------------------
    
    rng     = np.random.default_rng()
    n       = len(gate_pool) + 1
    c_range = np.arange(n-1)
          
    tensor = []
    label  = []
    U      = []
    
    for i in tqdm(range(samples), disable=silent):
        sub_gate_pool_cnt = rng.integers(min_sub_gate_pool_cnt, n)       
        sub_gate_pool_ind = rng.choice(c_range, size=sub_gate_pool_cnt, replace=False)         
        sub_gate_pool     = [gate_pool[ind] for ind in sub_gate_pool_ind]  # pick random subeset of gates
              
        qc, qc_tensor, svr = get_rnd_encoded_circuit(num_of_qubits, min_gates, max_gates, sub_gate_pool, gate_classes, rng, optimized=True)
    
        tensor.append(qc_tensor)
        label.append(f"Compile using: {[x().name for x in sub_gate_pool]}")
        U.append(qi.Operator(qc).to_matrix())
        
    #-------------------------------
        
    # make sure we have unique circuits
    tensor = torch.stack(tensor, dim=0)
    tensor_unique, tensor_indices = dahe.get_unique_elements_indices(tensor)
    
    if not silent: print(f"generated unique circuits: {tensor_unique.shape[0]}")
    
    #--------------------------
    #select uniques only
    
    x = tensor[tensor_indices]
    y = [label[i] for i in tensor_indices.tolist()]
    U = [    U[i] for i in tensor_indices.tolist()]
    
    return x,y,U  



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/printing.ipynb.

# %% auto 0
__all__ = ['display_colums', 'ndarray_to_latex', 'tensor_to_latex', 'print_markdown', 'print_table']

# %% ../src/printing.ipynb 3
from .imports import *
from ipywidgets import widgets
if IN_NOTEBOOK: from IPython.display import Markdown

# %% ../src/printing.ipynb 4
def display_colums(display_list, num_col=3):
    
    outputs = [widgets.Output() for i in range(num_col)]

    for i in range(len(display_list)//num_col+1):
        
        ds = display_list[i*num_col:(i+1)*num_col]
   
        for d,output in zip(ds,outputs):
            with output:
                display(d)
             
    columns = widgets.HBox(outputs)
    display(columns)  

# %% ../src/printing.ipynb 8
def ndarray_to_latex(arr):
    """Returns a LaTeX `{pmatrix*}[r]` as a string"""
    if len(arr.shape) > 2: raise ValueError('pmatrix can at most display two dimensions')
    lines = str(arr).replace('[', '').replace(']', '').splitlines()
    rv = [r'\begin{pmatrix*}[r]']
    rv += ['  ' + ' & '.join(l.split()) + r'\\' for l in lines]
    rv +=  [r'\end{pmatrix*}']
    return '\n'.join(rv)

# %% ../src/printing.ipynb 9
def tensor_to_latex(tensor):
    """Returns a `LaTeX {pmatrix*}[r]` as a string """
    if len(tensor.shape) > 2: raise ValueError('pmatrix can at most display two dimensions')
    lines = str(tensor.numpy()).replace('[', '').replace(']', '').splitlines()
    rv = [r'\begin{pmatrix*}[r]']
    rv += ['  ' + ' & '.join(l.split()) + r'\\' for l in lines]
    rv +=  [r'\end{pmatrix*}']
    return '\n'.join(rv)

# %% ../src/printing.ipynb 11
def print_markdown(text, print_raw=False):
    if IN_NOTEBOOK and not print_raw: display(Markdown(text))
    else: print(text)

# %% ../src/printing.ipynb 14
def print_table(col_headings: list, data: np.array, row_headings=None, print_raw=False):   
    assert len(col_headings) == data.shape[1]
    if row_headings is not None: assert len(row_headings) == data.shape[0]
        
    #--------------------------------
    head = ""
    if row_headings is not None: head = "| " + head  
    
    for col_heading in col_headings: head += f"|{col_heading}"
    head += "|\n"
                
    #--------------------------------
    seperator = ""
    if row_headings is not None: seperator = "|--"
        
    for col_heading in col_headings: seperator += "|--"
    seperator += "|\n"
    
    #--------------------------------    
    body = ""
    for i, row in enumerate(data):
        body_row = ""
        for x in row:
            body_row += f"|{x:.2f}"
            
        if row_headings is not None: 
            body_row = f"|{row_headings[i]}" + body_row
        
        body += body_row + "|\n"
                
    #--------------------------------    
    table = head + seperator + body
      
    print_markdown(table, print_raw)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/unitary_encoder.ipynb.

# %% auto 0
__all__ = ['Unitary_encoder_config', 'Unitary_encoder']

# %% ../../src/models/unitary_encoder.ipynb 2
from ..imports import *
from .config_model import Config_Model
import genQC.models.layers as layers
import genQC.models.transformers as transformers

# %% ../../src/models/unitary_encoder.ipynb 4
@dataclass
class Unitary_encoder_config:  
    cond_emb_size: int
    model_features: list[int] 
    num_heads: int
    transformer_depths: list[int]
    dropout: float

# %% ../../src/models/unitary_encoder.ipynb 5
class Unitary_encoder(Config_Model):
    """Encoder for unitary conditions."""
    def __init__(self, cond_emb_size, model_features=None, num_heads=8, transformer_depths=[4, 4], dropout=0.1):
        super().__init__()             

        self.cond_emb_size = cond_emb_size
        
        if not exists(model_features):
            in_ch   = 2                  # complex splitted in real and img  
            mid_ch1 = cond_emb_size//4
            mid_ch2 = cond_emb_size//2
            out_ch  = cond_emb_size
            
            model_features = [in_ch, mid_ch1, mid_ch2, out_ch]

        else:
            assert len(model_features) == 4
            in_ch, mid_ch1, mid_ch2, out_ch = model_features
    
        #------------------------------------
        
        self.params_config = Unitary_encoder_config(cond_emb_size, model_features, num_heads, transformer_depths, dropout)

        #------------------------------------
          
        self.conv_in = nn.Conv2d(in_ch, mid_ch1, kernel_size=1, stride=1, padding ="same")
        self.pos_enc = layers.PositionalEncoding2D(d_model=mid_ch1) 

        self.down1 = layers.DownBlock2D(mid_ch1, mid_ch2, kernel_size=(2, 2), stride=(2, 2), padding=(0,0))    

        #------------
        assert len(transformer_depths) == 2        
        self.spatialTransformer1 = transformers.SpatialTransformerSelfAttn(mid_ch1, num_heads=num_heads, depth=transformer_depths[0], dropout=dropout)
        self.spatialTransformer2 = transformers.SpatialTransformerSelfAttn(mid_ch2, num_heads=num_heads, depth=transformer_depths[1], dropout=dropout)

        #------------
        self.head = nn.Conv2d(mid_ch2, out_ch, kernel_size=1, stride=1, padding ="same")    

        #------------------------------------
        
        self._init_weights()
    
    def _init_weights(self):
        self.head.weight.data.zero_()
    
    def forward(self, x): 
        # x ... [batch, 2, 2^n, 2^n]     n=num_of_qubits
        
        b, *_ = x.shape
        
        x = self.conv_in(x)
        x = self.pos_enc(x)     
        
        x = self.spatialTransformer1(x)
        x = self.down1(x)
                
        x = self.spatialTransformer2(x)
        
        #-------------------
        x = self.head(x)
        x = torch.reshape(x, (b, self.cond_emb_size, -1)) # [batch, ch, x, y] to  [batch, ch, seq]
        x = torch.permute(x, (0, 2, 1))                   # [batch, ch, seq]  to [batch, seq, ch]
              
        return x  



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/dataset/qc_dataset.ipynb.

# %% auto 0
__all__ = ['Qc_Config_Dataset_config', 'Qc_Config_Dataset']

# %% ../../src/dataset/qc_dataset.ipynb 3
from ..imports import *
from .config_dataset import Config_Dataset, Config_Dataset_config
from ..config_loader import *
from .dataset_helper import *
from ..platform.qcircuit_dataset_construction import decode_circuit
from ..platform.simulation.qcircuit_sim import schmidt_rank_vector, instruction_name_to_qiskit_gate
import qiskit.quantum_info as qi

# %% ../../src/dataset/qc_dataset.ipynb 4
@dataclass
class Qc_Config_Dataset_config(Config_Dataset_config):
    optimized: bool
    dataset_to_gpu: bool
    random_samples: int  
    num_of_qubits: int  
    min_gates: int 
    max_gates: int 
    gate_pool: list[str] 

# %% ../../src/dataset/qc_dataset.ipynb 5
class Qc_Config_Dataset(Config_Dataset):
    """Dataset for quantum circuits, access `gate_pool` directly and all other paras with `.params_config`"""
    
    req_params     = [f.name for f in dataclasses.fields(Qc_Config_Dataset_config)]
    add_balance_fn = None
    
    def __init__(self, device: torch.device=torch.device("cpu"), **parameters):
        super().__init__(device, **parameters)            
        self.gate_pool = parameters["gate_pool"] #[get_obj_from_str(gate) for gate in parameters["gate_pool"]]           
            
    @property
    def params_config(self):
        params_config = super().params_config            
        #params_config["gate_pool"] = [class_to_str(gate) for gate in params_config["gate_pool"]]
        params_config = Qc_Config_Dataset_config(**params_config)
        return params_config   
    
    #----------------------------
       
    def x_y_preprocess(self, balance_max=None, shuffle=False, max_samples=None):
        #params_config = self.params_config
        #if params_config.dataset_to_gpu: self.to("cuda")
         
        z_proc = []
        for k,v in self.store_dict.items(): 
            if k != "x" and k != "y":
                z_proc.append(getattr(self, k))
                              
        x_proc, y_proc = self.x, self.y
        
        #---------------------
        if shuffle:
            x_proc, y_proc, *z_proc = shuffle_tensor_dataset(x_proc, y_proc, *z_proc)
                
        if exists(max_samples):
            x_proc = x_proc[:max_samples]
            y_proc = y_proc[:max_samples]
            z_proc = (iz[:max_samples] for iz in z_proc)   
        
        #---------------------
        t = self.store_dict["y"]
        if exists(balance_max): 
            if t == "tensor" or t == "numpy": x_proc, y_proc, *z_proc = balance_tensor_dataset(x_proc, y_proc, *z_proc, make_unique=True, 
                                                                                               samples=balance_max, add_balance_fn=self.add_balance_fn) 
            else:                             print(f"[WARNING]: Unsupported y type: `{t}`. Not balancing dataset!")
        else: print(f"[INFO]: Not balancing dataset!  {balance_max=}")
          
        #---------------------
        if shuffle:
            x_proc, y_proc, *z_proc = shuffle_tensor_dataset(x_proc, y_proc, *z_proc)
            
        return x_proc, y_proc, *z_proc
    
    def valid_split(self, x, y, *z, p_valid=0.1):
        splits = max(int(x.shape[0] * p_valid), 1)
        x, x_valid =    x[splits:].clone(), x[:splits].clone() 
        
        t = self.store_dict["y"]
        if   t == "tensor" : y, y_valid =    y[splits:].clone(), y[:splits].clone()      
        elif t == "numpy":   y, y_valid =    y[splits:]        , y[:splits]      
            
        else: raise NotImplementedError("Not implemented")
        
        try:
            z       = list(iz[splits:].clone() for iz in z)
            z_valid = list(iz[:splits].clone() for iz in z)     
        except:
            z       = list(iz[splits:] for iz in z)
            z_valid = list(iz[:splits] for iz in z)   
        
        return x, x_valid, y, y_valid, (z, z_valid)
    
    def get_dataloaders(self, batch_size, p_valid=0.1, balance_max=None, max_samples=None, y_on_cpu=False):
 
        excepts = []
        if y_on_cpu: excepts.append("y")
        if self.params_config.dataset_to_gpu: self.to("cuda", excepts=excepts)
               
        x_proc, y_proc, *z_proc              = self.x_y_preprocess(balance_max=balance_max, max_samples=max_samples)       
        x, x_valid, y, y_valid, (z, z_valid) = self.valid_split(x_proc, y_proc, *z_proc, p_valid=p_valid)
               
        ds       = TensorDataset(x, y, *z)
        ds_valid = TensorDataset(x_valid, y_valid, *z_valid)
        
        if self.params_config.dataset_to_gpu: 
            train_loader = DataLoader(dataset=ds      , batch_size=batch_size, shuffle=True)
            valid_loader = DataLoader(dataset=ds_valid, batch_size=batch_size, shuffle=True)

        else:              
            train_loader = DataLoader(dataset=ds      , batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=12)
            valid_loader = DataLoader(dataset=ds_valid, batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=12)

        self.dataloaders = DataLoaders(train_loader, valid_loader)        
        return self.dataloaders

    #----------------------------
  
    def plot_example(self):
        params_config = self.params_config
        enc_tensor = self.x[0]
        
        while enc_tensor.dim()>2: enc_tensor=enc_tensor[0]

        params = None
        if hasattr(self, "params"): params=self.params[0]

        if isinstance(self.gate_pool[0], str):
            gate_pool = [instruction_name_to_qiskit_gate(gate) for gate in self.gate_pool]
        else:
            gate_pool = self.gate_pool
        
        qc = decode_circuit(enc_tensor, gate_pool, params_tensor=params)
 
        t = self.store_dict["y"]
        if   t == "tensor"     : label = self.y[0].cpu().tolist()
        elif t == "tensor_list": 
            print("Not implemented")
            return
        else : 
            label = self.y[0]#.tolist()
            while len(label.shape)>0: label=label[0]
        
        print(f"Label: ``{label}``    SRV is: {schmidt_rank_vector(qi.DensityMatrix(qc))}")
        display(qc.draw("mpl", plot_barriers=False))

    def plot_distribution(self):
        if hasattr(self, "dataloaders"): x, y, *z = self.dataloaders.train.dataset.tensors  
        else:                            x, y     = self.x, self.y
            
        t = self.store_dict["y"]
        if   t == "tensor"     : data={"svr":[iy for iy in y.cpu().tolist()]}
        elif t == "numpy": data={"svr":[iy for iy in y.tolist()]}
        else:   #  list tensor_list        
            print("Not implemented")
            return
                        
        print("Train dataset (x, y):", x.shape, y.shape)
        print("Train uniques  x    :", torch.unique(x, dim=0).shape)              
              
        #real data distribution     
        df = pd.DataFrame(data)   
        cnts = df['svr'].value_counts(normalize=True)
        for n,v in zip(cnts.index, cnts.values): print(f"{n}: {v*100:.1f}%")    
        ax = df['svr'].value_counts().plot(kind='bar')



###################################
## EDIT HERE, NOT AUTOGENERATED! ## 
###################################
#------------------------------------
# Python

import math, itertools, functools, copy, asyncio, time, importlib, datetime, importlib, os, dataclasses, platform
from datetime import datetime
from PIL import Image
from dataclasses import dataclass, asdict, is_dataclass
from typing import Union, Optional, TypeVar, Callable, Any

import numpy as np
import pandas as pd
import scipy
import matplotlib.pyplot as plt

#------------------------------------
# PyTorch

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torchvision import datasets
from torch.utils.data import DataLoader, TensorDataset

torch.set_printoptions(linewidth=200)

#------------------------------------
# runtime

def in_colab():
    "Check if the code is running in Google Colaboratory"
    try:
        from google import colab
        return True
    except: return False

def in_notebook():
    "Check if the code is running in a jupyter notebook"
    if in_colab(): return True
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell': # Jupyter notebook, Spyder or qtconsole
            import IPython
            #IPython version lower then 6.0.0 don't work with output you update
            return IPython.__version__ >= '6.0.0'
        elif shell == 'TerminalInteractiveShell': return False  # Terminal running IPython
        else: return False  # Other type (?)
    except NameError: return False      # Probably standard Python interpreter

IN_NOTEBOOK = in_notebook()

if IN_NOTEBOOK: from tqdm.notebook import trange, tqdm
else: from tqdm import trange, tqdm

#------------------------------------
# python commons

def exists(val): return val is not None
def default(val, d):
    if exists(val): return val
    return d() if isfunction(d) else d

#------------------------------------
# ....



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/frozen_open_clip.ipynb.

# %% auto 0
__all__ = ['FrozenOpenCLIPEmbedder_config', 'FrozenOpenCLIPEmbedder', 'CachedFrozenOpenCLIPEmbedder']

# %% ../../src/models/frozen_open_clip.ipynb 2
from ..imports import *
from .config_model import Config_Model
import open_clip

# %% ../../src/models/frozen_open_clip.ipynb 4
@dataclass
class FrozenOpenCLIPEmbedder_config:
    arch: str
    version: str
    device: str
    max_length: int
    freeze: bool
    layer: str

# %% ../../src/models/frozen_open_clip.ipynb 5
class FrozenOpenCLIPEmbedder(Config_Model):
    """Loads and freezes the [OpenCLIP](https://github.com/mlfoundations/open_clip) transformer encoder for text prompts."""
    
    LAYERS = [
        # "pooled",
        "last",
        "penultimate"
    ]

    def __init__(self, arch="ViT-H-14", version="laion2b_s32b_b79k", device="cpu", max_length=77, freeze=True, layer="penultimate"):
        super().__init__()               
        assert layer in self.LAYERS     
        self.params_config = FrozenOpenCLIPEmbedder_config(arch, version, device, max_length, freeze, layer)
        
        model, _, _ = open_clip.create_model_and_transforms(arch, device=torch.device(device), pretrained=version)
        del model.visual
        
        self.model = model
        self.to(device)

        self.tokenizer = open_clip.get_tokenizer(arch)
        
        assert max_length <= 77   # max set by the clip 
        self.max_length = max_length
        
        if freeze: self.freeze()
        
        self.layer = layer
        if   self.layer == "last":         self.layer_idx = 0
        elif self.layer == "penultimate":  self.layer_idx = 1
        else: raise NotImplementedError()

        #create empty token, can also be, e.g., A nice picture
        self.empty_token = self.tokenize_and_push_to_device("")
        
    def freeze(self):
        self.model = self.model.eval()
        
        for param in self.parameters(): 
            param.requires_grad = False    
            
        for param in self.model.parameters(): 
            param.requires_grad = False
    
    def to(self, device):
        self.model  = self.model.to(device)           
        self.device = device
        return self
        
    @torch.no_grad()
    def tokenize_and_push_to_device(self, text, to_device=True):
        # tokens = open_clip.tokenize(text)
        tokens = self.tokenizer(text)
        if to_device:
            tokens = tokens.to(self.device)
        return tokens
    
    @torch.no_grad()
    def forward(self, c, **kwargs):
        return self.encode_with_transformer(c)

    @torch.no_grad()
    def encode_with_transformer(self, text):
        cast_dtype = self.model.transformer.get_cast_dtype()
        
        x = self.model.token_embedding(text).to(cast_dtype)  # [batch_size, n_ctx, d_model]        
        x = x + self.model.positional_embedding[None, :x.shape[1]].to(cast_dtype)

        if not self.model.transformer.batch_first:
            x = x.permute(1, 0, 2)  # NLD -> LND
        
        x = self.text_transformer_forward(x, attn_mask=self.model.attn_mask)

        if not self.model.transformer.batch_first:
            x = x.permute(1, 0, 2)  # LND -> NLD
        
        x = self.model.ln_final(x)  # [batch_size, n_ctx, transformer.width]
        
        return x

    @torch.no_grad()
    def text_transformer_forward(self, x: torch.Tensor, attn_mask=None):
        for i, r in enumerate(self.model.transformer.resblocks):
            if i == len(self.model.transformer.resblocks) - self.layer_idx:
                break

            x = r(x, attn_mask=attn_mask)  
        return x

    #--------------------------------------------------------------
    
    def get_config(self, save_path=None, without_metadata=False):
        return super().get_config(save_path=None, without_metadata=without_metadata)
    
    def store_model(self, config_path: str, save_path: str=None, without_metadata=False):        
        super().store_model(config_path, save_path=None, without_metadata=without_metadata)

    @staticmethod
    def from_config(config, device: torch.device, save_path: str=None):  
        config["save_path"] = None
        return Config_Model.from_config(config, device, save_path=None)        

# %% ../../src/models/frozen_open_clip.ipynb 17
class CachedFrozenOpenCLIPEmbedder(FrozenOpenCLIPEmbedder):
    """Adds caching support to `FrozenOpenCLIPEmbedder`."""
    
    @torch.no_grad()
    def generate_cache(self, str_list: list=None, tokens=None, cached_empty_token_index=0, b_size=2048, y_on_cpu=False):       
        self.cached_empty_token_index = cached_empty_token_index       
        if exists(str_list): self.cached_tokens = self.tokenize_and_push_to_device(str_list)      
        elif exists(tokens): self.cached_tokens = tokens
        else: raise RuntimeError("please provide str_list or tokens")
        
        # note: we need to split the tokens in batches for forward pass, n gets large
        # cached_tokens     [n, 77]      ... int
        # cached_embeddings [n, 77, 512] ... float

        n = self.cached_tokens.shape[0]
        
        n_chunks = int(np.ceil(n / b_size))
        
        in_device = self.cached_tokens.device
                
        last_ind = 0
        for i, cached_tokens in tqdm(enumerate(self.cached_tokens.chunk(n_chunks)), total=n_chunks):
            
            x = super().forward(cached_tokens.to(self.device))
            
            if i == 0:
                mem = n * x.shape[1] * x.shape[2] * x.element_size() * 1e-9
                print(f"[INFO]: caching trying to allocate memory {(n, x.shape[1], x.shape[2])} on {'cpu' if y_on_cpu else self.device}, approx. {mem:.3f} GB")
                self.cached_embeddings = torch.zeros((n, x.shape[1], x.shape[2]), device="cpu" if y_on_cpu else self.device, dtype=x.dtype) # alloc huge memory !!
                
            self.cached_embeddings[last_ind:last_ind+x.shape[0]] = x.to(self.cached_embeddings.device)
            
            last_ind += x.shape[0]
            
        if not y_on_cpu:
            self.cached_embeddings = self.cached_embeddings.to(in_device)

    @torch.no_grad()
    def look_up_cos_sim_cached_index(self, str_list: list=None, tokens=None):
        if exists(str_list): tokens = self.tokenize_and_push_to_device(str_list)      
        else: raise RuntimeError("please provide str_list or tokens")
                                         
        emb   = super().forward(tokens.to(self.device))
        c_emb = self.cached_embeddings
        #-----------------
        # do cos sim search
        
        emb   = emb.flatten(start_dim=1)   # [m, seq*ch]
        c_emb = c_emb.flatten(start_dim=1) # [n, seq*ch]

        norm_emb   =   emb / torch.linalg.vector_norm(  emb, dim=1, keepdim=True)
        norm_c_emb = c_emb / torch.linalg.vector_norm(c_emb, dim=1, keepdim=True) 
 
        sim     = torch.matmul(norm_c_emb, norm_emb.T) # matmul out is [n, m]
        max_idx = torch.argmax(sim, dim=0)             # reduce the c_emb dim, [m]
     
        return max_idx       
                            
    @torch.no_grad()
    def forward(self, c, **kwargs):  
        in_device = c.device
        
        if   c.dim() == 1: return self.cached_embeddings[c.to(self.cached_embeddings.device)].to(in_device)         #list of ints       
        elif c.dim() == 2: return super().forward(c, **kwargs)   #tokenized input      
        else: raise NotImplementedError("")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/dataset/config_dataset.ipynb.

# %% auto 0
__all__ = ['Config_Dataset_config', 'Config_Dataset']

# %% ../../src/dataset/config_dataset.ipynb 2
from ..imports import *
from ..config_loader import *

from huggingface_hub import snapshot_download

# %% ../../src/dataset/config_dataset.ipynb 3
@dataclass
class Config_Dataset_config:
    """Config `dataclass` used for storage."""
    store_dict: dict 

# %% ../../src/dataset/config_dataset.ipynb 4
class Config_Dataset():  
    """Base class for datasets, manages loading and saving."""
    
    req_params = [f.name for f in dataclasses.fields(Config_Dataset_config)]
    comment    = ""
    
    def __init__(self, device: torch.device=torch.device("cpu"), **parameters):
        req_params = self.req_params      
        for p in req_params:
            if p not in parameters: raise RuntimeError(f"Missing parameter `{p}` in argument `**parameters: dict`")           

        self.device = device      #parameters will overwrite passed device
        
        for k,v in parameters["store_dict"].items(): 
            if   v == "tensor"     : setattr(self, str(k), torch.tensor([0], device=self.device))
            elif v == "tensor_list": setattr(self, str(k), [torch.tensor([0], device=self.device)]) 
            elif v == "list"       : setattr(self, str(k), ["list str entry"]) 
            elif v == "numpy"      : setattr(self, str(k), np.array(["numpy str entry"]))                                             
            else                   : raise RuntimeError(f"Unknown type `{v}` in argument parameters[`store_dict`]")
                                             
        for k,v in parameters.items(): setattr(self, str(k), v)
           
    def to(self, device: torch.device, excepts=[], **kwargs):
        self.device = device
        
        for k,v in self.store_dict.items(): 
            if k in excepts: continue
            
            if v == "tensor": 
                x = getattr(self, str(k)).to(device, **kwargs)
                setattr(self, str(k), x)
                
            elif v == "tensor_list": 
                x = getattr(self, str(k))
                x = [ix.to(device, **kwargs) for ix in x]
                setattr(self, str(k), x)
  
        return self
    
    #----------------------------
    
    @property
    def params_config(self):
        params_config = {}              
        for p in self.req_params: params_config[p] = getattr(self, p)
        return params_config   
       
    def get_config(self, save_path=None, without_metadata=False):
        if not without_metadata:       
            config = {}
            config["target"]         = class_to_str(type(self))
            config["device"]         = str(self.device)
            config["comment"]        = self.comment
            config["save_path"]      = self.save_path if hasattr(self, "save_path") else save_path
            config["save_datetime"]  = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
            config["params"]         = self.params_config  
        else:
            config = self.params_config  
        
        self.config = config        
        return config
            
    def save_dataset(self, config_path: str, save_path: str):
        config = self.get_config(save_path, without_metadata=False)
        save_dict_yaml(config, config_path)                                   
        self.store_x_y(save_path)          
   
    #----------------------------
    
    def store_x_y(self, path_str):       
        for k,v in self.store_dict.items(): 
            x = getattr(self, str(k))
            torch.save(x, path_str + f"_{k}.pt")
                
    def load_x_y(self, path_str):
        self.save_path = path_str
        
        for k,v in self.store_dict.items(): 
            x = torch.load(path_str + f"_{k}.pt", weights_only=False)
            setattr(self, str(k), x)
        
    #----------------------------
    
    @staticmethod
    def from_config(config, device: torch.device, save_path: Optional[str] = None, make_contiguous: bool = True):
        """Use this if we have a loaded config."""
        
        config_dataset = instantiate_from_config(config)
        
        if "comment" in config: config_dataset.comment = config["comment"]
        
        #--------------------------------  

        config_dataset.save_type = config.pop("save_type", None)
        
        if not exists(save_path):            
            if "save_path" in config: save_path = config["save_path"]
            else:                     print("[INFO]: Found no key `save_path` path in config and no `save_path` arg provided.")
                                  
        if exists(save_path): config_dataset.load_x_y(save_path)
        else:                 print("[INFO]: No save_path` provided. Nothing loaded.")

        #--------------------------------
        
        config_dataset = config_dataset.to(device)
        print(f"[INFO]: Instantiated config_dataset from given config on {device}.")
        
        return config_dataset
    
    @classmethod
    def from_config_file(cls, config_path, device: torch.device, save_path:  Optional[str] = None, make_contiguous: bool = True):
        """
        Load a dataset from a config file. 
        If this method is called with `ConfigDataset.from_config_file` we use the given `target`, else use the caller class.
        """
        config = load_config(config_path)
        if cls is not Config_Dataset:
            config["target"] = class_to_str(cls)      
        return cls.from_config(config, device, save_path, make_contiguous)

    @classmethod
    def from_huggingface(cls, repo_id: str, device: torch.device, **kwargs):  
        """Load a dataset directly from Huggingface."""
        dataset_path = snapshot_download(repo_id=repo_id, repo_type="dataset", allow_patterns=["*.pt", "*.yaml", "*.safetensors"], **kwargs) 
        dataset = cls.from_config_file(config_path=dataset_path+"/config.yaml", device=device, save_path=dataset_path+"/dataset")  
        return dataset 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/platform/qcircuit_util.ipynb.

# %% auto 0
__all__ = ['get_element_matching_indices', 'get_entanglement_bins']

# %% ../../src/platform/qcircuit_util.ipynb 2
from ..imports import *

# %% ../../src/platform/qcircuit_util.ipynb 4
def get_element_matching_indices(a, b):
    """Compares (2d) `a` with `b`. Returns the indices of `b`, where a element of `a` matches with `b`."""
    # Expand dimensions of a to match the shape of b for element-wise comparison
    expanded_a = a.unsqueeze(0).expand(b.shape[0], *a.shape)  # [b0, a0, a1]
    expanded_b = b.unsqueeze(1)                               # [b0,  1, b1]
    
    # Compare all vector entries of a with all vectors of b
    matches = torch.all(expanded_a == expanded_b, dim=-1)

    matching_indices = torch.nonzero(torch.any(matches, dim=1)).squeeze()
    
    if   matching_indices.dim() == 0: matching_indices = torch.tensor([matching_indices])

    return matching_indices

# %% ../../src/platform/qcircuit_util.ipynb 5
def get_entanglement_bins(num_of_qubits):
    """Returns all SRV sorted in entangle bins which correspond to a number of entangled qubits."""
    dist_srvs = [x for x in itertools.product(*([[1,2]]*num_of_qubits))]
    dist_srvs = np.array(dist_srvs, dtype=int)[np.sum(dist_srvs, axis=1)!=num_of_qubits+1].tolist()
    dist_srvs = sorted(dist_srvs, key=lambda x: sum(x))
    dist_srvs = np.array(dist_srvs)
    
    entangle = [1] + [scipy.special.comb(num_of_qubits, i, exact=True) for i in range(2, num_of_qubits)]
    
    entanglement_bins = np.split(dist_srvs, np.cumsum(entangle))
    
    ent_bits = [f"{sum(n[0])-num_of_qubits} qubit entangled" for n in entanglement_bins]
    
    return [x.tolist() for x in entanglement_bins], ent_bits



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/pipeline/pipeline.ipynb.

# %% auto 0
__all__ = ['Pipeline_IO', 'Pipeline']

# %% ../../src/pipeline/pipeline.ipynb 3
from ..imports import *
from ..util import virtual, number_of_paramters, DataLoaders
from ..metrics import *
from ..config_loader import *

# %% ../../src/pipeline/pipeline.ipynb 5
nn.Loss = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]

# %% ../../src/pipeline/pipeline.ipynb 8
class Pipeline_IO:   
    """A class providing basic IO functionality."""
    def get_config(self, save_path: str, without_metadata=False):       
        params_config = self.params_config(save_path)  
              
        if not without_metadata:       
            config = {}
            config["target"]         = class_to_str(type(self))
            config["save_datetime"]  = datetime.now().strftime("%m/%d/%Y %H:%M:%S")
            config["params"]         = params_config             
            
            fit = {}
            if hasattr(self, "num_epochs"):         fit["num_epochs"]         = self.num_epochs
            if hasattr(self, "batch_size"):         fit["batch_size"]         = self.batch_size
            if hasattr(self, "lr"):                 fit["lr"]                 = self.lr
            if hasattr(self, "lr_sched"):           fit["lr_sched"]           = class_to_str(type(self.lr_sched))  
            if hasattr(self, "optimizer"):          fit["optimizer"]          = class_to_str(type(self.optimizer))   
            if hasattr(self, "dataset_size_train"): fit["dataset_size_train"] = self.dataset_size_train
            if hasattr(self, "dataset_size_valid"): fit["dataset_size_valid"] = self.dataset_size_valid
            config["fit"] = fit 
            
        else:
            config = params_config
            
        self.config = config        
        return config
    
    @virtual
    def params_config(self, save_path: str): return None
    
    def store_pipeline(self, config_path: str, save_path: str): 
        if exists(config_path): os.makedirs(config_path, exist_ok=True)       
        if exists(save_path): 
            os.makedirs(save_path, exist_ok=True)
            if hasattr(self, "fit_losses"):       np.savetxt(save_path + "fit_losses.txt", self.fit_losses)            
            if hasattr(self, "fit_valid_losses"): np.savetxt(save_path + "fit_valid_losses.txt", self.fit_valid_losses) 
                             
    @virtual
    @staticmethod
    def from_config_file(config_path, device: torch.device, save_path: str=None): return None 

# %% ../../src/pipeline/pipeline.ipynb 9
class Pipeline(Pipeline_IO):
    """A `Pipeline_IO` class providing basic pytorch model training functionality."""
    def __init__(self, 
                 model: nn.Module,
                 device: torch.device):
        self.model  = model
        self.device = device
      
    #------------------------------------
             
    @virtual
    def __call__(self, inp): pass
        
    @virtual
    def train_step(self, data, train=True, **kwargs): pass

    #------------------------------------
     
    def compile(self, optim_fn: type(torch.optim.Optimizer), loss_fn: nn.Loss, metrics: Union[Metric, list[Metric]]=None, lr=None, **kwargs):       
        self.loss_fn    = loss_fn()
        self.optim_fn   = optim_fn 
        self.optimizer  = optim_fn(self.model.parameters(), lr=lr, **kwargs) if lr else None
        
        metrics = {m.name:m for m in metrics} if metrics else {}
        #metrics |= {f"{m.name}_valid":m for m in metrics.values()}
        metrics["loss"]       = Mean("loss", self.device)  
        metrics["loss_valid"] = Mean("loss_valid", self.device)  
                          
        self.metrics = metrics    
                          
    def _reset_opt(self, lr, **kwargs): self.optimizer = self.optim_fn(self.model.parameters(), lr, **kwargs)   
    
    def _set_opt_param(self, lr, **kwargs):
        '''at least lr: Does not reset existing optimizer, only changes learn rate.'''
        self.lr = lr
        if lr:  
            if self.optimizer:           
                for g in self.optimizer.param_groups: 
                    g['lr'] = lr
                    for k,v in kwargs.items(): g[k] = v                  
            else: self._reset_opt(lr, **kwargs)
         
    #------------------------------------

    def train_on_batch(self, data, train=True):
        loss = self.train_step(data, train=train)   
          
        if train:            
            #zero grads
            self.optimizer.zero_grad()

            #backprob
            loss.backward()

            #update weights
            self.optimizer.step()
    
        return loss.detach()
    
    def train_on_epoch(self, data_loader: DataLoader, train=True):   
        self.model.train(train)
        
        mode = "" if train else "_valid"
        
        with self.progress_bar(total=len(data_loader), epoch=self.epoch, unit=" batch") as batch_prgb:                   
            for batch, data in enumerate(data_loader):    
                loss = self.train_on_batch(data, train=train)                                                                       
                self.metrics["loss"+mode].update_state(loss)            
                
                if train:
                    self.fit_losses.append(loss.item())                
                    if self.lr_sched: self.lr_sched.step()
                
                #pack up metrics
                self.out_metric_dict = {m.name:m.result().tolist() for m in self.metrics.values() if not m.empty}               
                self.end_batch_metrics(batch_prgb, **self.out_metric_dict)   
        
    #run on train and one on valid
    def fit(self, num_epochs: int, data_loaders: DataLoaders, lr: float=None, lr_sched=None, log_summary=True):
        if not hasattr(self, "loss_fn"): raise RuntimeError("'compile' has to be called first")       
       
        self._set_opt_param(lr=lr)    
        if lr_sched: self.lr_sched = lr_sched(self.optimizer)
        else: self.lr_sched = None
            
        self.num_epochs = num_epochs
        self.epochs     = range(num_epochs)
        self.fit_losses = []
        self.fit_valid_losses = []                 
        self.batch_size = data_loaders.train.batch_size 
        self.dataset_size_train = len(data_loaders.train)
        if data_loaders.valid: self.dataset_size_valid = len(data_loaders.valid)
        
            
        with self.progress_bar(total=num_epochs, desc="Fit", unit=" epoch") as epoch_prgb:       
            for self.epoch in self.epochs:   
                        
                #reset all metrics
                for m in self.metrics.values(): m.reset_state()       
                
                #train set
                self.train_on_epoch(data_loaders.train, train=True) 
          
                #valid set
                if data_loaders.valid: 
                    self.train_on_epoch(data_loaders.valid, train=False) 
                    self.fit_valid_losses.append([(self.epoch+1)*len(data_loaders.train), 
                                                   self.out_metric_dict["loss_valid"] ]) 
                    
                self.end_epoch_metrics(epoch_prgb, **self.out_metric_dict)

        self.fit_summary(log_summary=log_summary)
                
    #------------------------------------
                       
    def summary(self): print("Number of model parameters:", number_of_paramters(self.model))

    def fit_summary(self, figsize=(12,2), log_summary=True, return_fig=False):
        fig = plt.figure(figsize=figsize, constrained_layout=True)                
        plt.xlabel("Batches")
        plt.ylabel("Loss")
        if log_summary: plt.yscale('log') 
        plt.plot(self.fit_losses, label="train")
        if len(self.fit_valid_losses) > 0: 
            data = np.array(self.fit_valid_losses)
            plt.plot(data[:,0],data[:,1], label="valid", color="tab:orange")
            plt.plot(data[:,0],data[:,1], ".", color="tab:orange")
        plt.legend()
        if return_fig: return fig
        plt.show()
             
    #------------------------------------
        
    def progress_bar(self, iterable=None, total=None, epoch: int=None, **progress_bar_config): 
        if not hasattr(self, "_progress_bar_config"):
            self._progress_bar_config = {}
        elif not isinstance(self._progress_bar_config, dict):
            raise ValueError(f"`self._progress_bar_config` should be of type `dict`, but is {type(self._progress_bar_config)}.")
        
        prgb_conf = self._progress_bar_config 
        
        if epoch is not None: prgb_conf |= {"desc":f"Epoch {epoch}"} 
        if progress_bar_config is not None: prgb_conf |= progress_bar_config  

        if iterable is not None: return tqdm(iterable, **self._progress_bar_config)
        elif total is not None:
            self._n_total = total
            return tqdm(total=total, **self._progress_bar_config)
        else: raise ValueError("Either `total` or `iterable` has to be defined.")
        
    def end_progress_bar_iteration(self, prgb:tqdm, print_lines=False, name="", index=None, **metrics):
        if metrics is not None: prgb.set_postfix(**metrics)        
        prgb.update()
        
        if not print_lines: return
        
        n_total = f"/{self._n_total}: " if hasattr(self, "_n_total") else ": "

        if index is not None: prgb.write(f"{name} {index:03}" + n_total, end='') 
        else: prgb.write(f"{name} {(prgb.n):03}" + n_total, end='') #(prgb.n+1)-1 bcs we update first

        if metrics is not None: prgb.write(str(metrics))            
        else: prgb.write("")  
                            
    def end_epoch_metrics(self, prgb:tqdm, epoch: int=None, **metrics): self.end_progress_bar_iteration(prgb, False, "Epoch", epoch, **metrics)             
    def end_batch_metrics(self, prgb:tqdm, batch: int=None, **metrics): self.end_progress_bar_iteration(prgb, False, "Batch", batch, **metrics)
    
    #------------------------------------
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/inference/export_cudaq.ipynb.

# %% auto 0
__all__ = ['backend', 'CircuitInstruction', 'CircuitInstructions', 'CircuitsCudaqBackend', 'tensor_to_instructions',
           'genqc_to_cudaq']

# %% ../../src/inference/export_cudaq.ipynb 2
from ..imports import *
from typing import Sequence, List, Optional
import cudaq

# %% ../../src/inference/export_cudaq.ipynb 4
@dataclass
class CircuitInstruction():
    name: str
    control_nodes: Sequence[int]
    target_nodes: Sequence[int]
    params: Sequence[float]

# %% ../../src/inference/export_cudaq.ipynb 5
class CircuitInstructions():
    def __init__(self, tensor_shape: torch.Size) -> None:
        assert len(tensor_shape) == 2   # ... [qubits, time]
        self.tensor_shape  = tensor_shape     
        self._instructions = []
        self.instruction_names_set = set()
    
    def add_instruction(self,  
                        name: str, 
                        control_nodes: Sequence[int], 
                        target_nodes: Sequence[int], 
                        params: Sequence[float]) -> None:
        self.instruction_names_set.add(name)
        self._instructions.append(CircuitInstruction(name, control_nodes, target_nodes, params))

    @property
    def data(self) -> List[CircuitInstruction]: return self._instructions

    @property
    def length(self) -> int: return len(self._instructions)

    @property
    def num_qubits(self) -> int: return self.tensor_shape[0]

    @property
    def max_gates(self) -> int: return self.tensor_shape[1]
    
    def __repr__(self) -> str: return str(self._instructions)

    def print(self) -> None:
        for instruction in self.data: 
            print(instruction)        

# %% ../../src/inference/export_cudaq.ipynb 7
class CircuitsCudaqBackend():

    BASIC_BACKEND_TYPE = type[cudaq.kernel]
    
    # Has to match with insides of belows kernel
    KERNEL_VOCABULARY = {"h":1, "cx":2, "z":3, "x":4, "y":5, "ccx":6, "swap":7} 
    
    def _construct_kernel(self,
                          gate_list: List[str],
                          target_1_nodes_list: List[int],
                          target_2_nodes_list: List[int],
                          control_1_nodes_list: List[int],
                          control_2_nodes_list: List[int]
                         ) -> cudaq.kernel:
        """Construct a `cudaq.kernel` from provided paramters."""
  
        num_gates = len(gate_list)
        gate_list = [self.KERNEL_VOCABULARY[g] for g in gate_list]

        # Note: `@cudaq.kernel` decorator has a overhead of 20ms, regardless of the for-loop inside
        
        @cudaq.kernel
        def place_gate_kernel(gate: int, 
                              qvector: cudaq.qview,
                              target_1: int, 
                              target_2: int, 
                              control_1: int, 
                              control_2: int):        
            if   gate == 1: h(qvector[target_1])
            elif gate == 2: cx(qvector[control_1], qvector[target_1])
            elif gate == 3: z(qvector[target_1])
            elif gate == 4: x(qvector[target_1])
            elif gate == 5: y(qvector[target_1])
            elif gate == 6: x.ctrl(qvector[control_1], qvector[control_2], qvector[target_1])
            elif gate == 7: swap(qvector[target_1], qvector[target_2])
      
        @cudaq.kernel  
        def kernel(input_state: List[complex]):
            qvector = cudaq.qvector(input_state)
            for i in range(num_gates):
                place_gate_kernel(gate_list[i], qvector, target_1_nodes_list[i], target_2_nodes_list[i], control_1_nodes_list[i], control_2_nodes_list[i])
    
        return kernel

    def check_error_circuit(self, 
                            gate: str, 
                            num_target_nodes: int, 
                            num_control_nodes: int) -> bool:
        """Check number of connections of given gate. Used to check for error circuits."""

        if gate not in self.KERNEL_VOCABULARY:
            raise NotImplementedError(f"Unknown gate {gate}, not in `self.KERNEL_VOCABULARY`.")
            
        if gate in ["h", "z", "x", "y"]:
            if num_target_nodes != 1 or num_control_nodes !=0: return False

        elif gate in ["cx"]:
            if num_target_nodes != 1 or num_control_nodes !=1: return False

        elif gate in ["ccx"]:
            if num_target_nodes != 1 or num_control_nodes !=2: return False

        elif gate in ["swap"]:
            if num_target_nodes != 2 or num_control_nodes !=0: return False

        else:
            raise NotImplementedError(f"Unknown gate {gate}, implemetation is faulty!")

        return True

    
    def export_cudaq(self, instructions: CircuitInstructions) -> cudaq.kernel:
        """Convert given genQC `CircuitInstructions` to a `cudaq.kernel`."""

        # num_qubits = instructions.num_qubits
        num_gates  = instructions.length

        # @cudaq.kernel can only take list[int] and no str directly
        # -> we have to map everything to list[int]        
        # set default value to 9999 so an error wil be raised if we have a faulty tensor encoding
        
        gate_list = []
        target_1_nodes_list  = [9999] * num_gates
        target_2_nodes_list  = [9999] * num_gates
        control_1_nodes_list = [9999] * num_gates
        control_2_nodes_list = [9999] * num_gates

        for i, instruction in enumerate(instructions.data):

            gate          = instruction.name.lower()
            control_nodes = instruction.control_nodes
            target_nodes  = instruction.target_nodes
 
            if len(instruction.params) > 0:
                raise NotImplementedError(f"Only support non parametrized gates currently.")
            
            num_target_nodes  = len(target_nodes)
            num_control_nodes = len(control_nodes)
            
            if not self.check_error_circuit(gate, num_target_nodes, num_control_nodes):
                return None
            
            gate_list.append(gate)
  
            if num_target_nodes > 0:
                target_1_nodes_list[i] = target_nodes[0]
                if num_target_nodes > 1: 
                    target_2_nodes_list[i] = target_nodes[1]      
            
            if num_control_nodes > 0:
                control_1_nodes_list[i] = control_nodes[0]  
                if num_control_nodes > 1: 
                    control_2_nodes_list[i] = control_nodes[1]  
                    
        #--------------------
        kernel= self._construct_kernel(gate_list, target_1_nodes_list, target_2_nodes_list, control_1_nodes_list, control_2_nodes_list)
        return kernel
    
    def get_unitary(self, kernel: cudaq.kernel, num_qubits: int) -> np.ndarray:
        """Return the unitary matrix of a `cudaq.kernel`. Currently relies on simulation, could change in future releases of cudaq."""
        
        N = 2**num_qubits
        U = np.zeros((N, N), dtype=np.complex128)
        
        for j in range(N): 
            state_j    = np.zeros((N), dtype=np.complex128) 
            state_j[j] = 1
            
            U[:, j] = np.array(cudaq.get_state(kernel, state_j), copy=False)
            
        return U

    def draw(self, kernel: cudaq.kernel, num_qubits: int, **kwargs) -> None:
        """Draw the given `cudaq.kernel` using cudaq.""" 
        c    = [0] * (2**num_qubits)
        c[0] = 1
        print(cudaq.draw(kernel, c))

# %% ../../src/inference/export_cudaq.ipynb 9
def tensor_to_instructions(tensor: torch.Tensor, 
                           vocabulary_inverse: dict, 
                           params_tensor: Optional[torch.Tensor] = None, 
                           params_4pi_normalization: bool = True,
                           sign_labels: dict = {"control_nodes":-1, "target_nodes":+1}) -> CircuitInstructions:
    """Convert a given `torch.Tensor` to `CircuitInstructions`."""
 
    assert tensor.dim() == 2, f"{tensor.shape=}"
    num_of_qubits, time = tensor.shape
    
    instructions = CircuitInstructions(tensor_shape=tensor.shape)
    
    for t in range(time):         
        enc_time_slice = tensor[:, t] # contains all bits at time t   
    
        for gate_index, gate in vocabulary_inverse.items():   
        
            target_nodes  = (enc_time_slice == (sign_labels["target_nodes"]  * gate_index)).nonzero(as_tuple=True)[0]
            control_nodes = (enc_time_slice == (sign_labels["control_nodes"] * gate_index)).nonzero(as_tuple=True)[0]
    
            if target_nodes.nelement() > 0:                                   
                params = []
                if exists(params_tensor):
                    params = params_tensor[:, t]
                    if params_4pi_normalization:
                        params = (params+1.0) * 2.0*np.pi    # [-1, 1] to [0, 4pi]
                    params = params.tolist()

                instructions.add_instruction(gate, control_nodes.tolist(), target_nodes.tolist(), params)
                
                break  #break on first hit, per def only one gate allowed per t
          
            elif control_nodes.nelement() > 0: # no target but control means error
                raise RuntimeError("target_nodes.nelement() <= 0 but control_nodes.nelement() > 0")
                
            #else we are fine with tensors that have time steps with no action!
    
    return instructions

# %% ../../src/inference/export_cudaq.ipynb 10
backend = CircuitsCudaqBackend()

def genqc_to_cudaq(tensor: torch.Tensor, vocabulary_inverse: dict) -> cudaq.kernel:
    """Convert given `torch.Tensor` to a `cudaq.kernel`."""
    instructions = tensor_to_instructions(tensor, vocabulary_inverse) 
    kernel       = backend.export_cudaq(instructions)
    return kernel



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/inference/infer_srv.ipynb.

# %% auto 0
__all__ = ['get_all_srvs', 'generate_srv_tensors', 'convert_tensors_to_srvs', 'get_srv_accuracy', 'true_sample_bin_dist',
           'test_srv_clr_distribution_bin_samples', 'test_srv_clr_distribution', 'test_guidance_dep',
           'test_srv_acc_vs_length', 'test_srv_acc_vs_maxLength', 'test_srv_length_distribution',
           'plot_srv_clr_distribution_hist', 'plot_srv_clr_distribution_bin_accuracy', 'plot_guidance_dep',
           'plot_srv_acc_vs_length', 'plot_srv_acc_vs_maxLength']

# %% ../../src/inference/infer_srv.ipynb 2
from ..imports import *
from .infer_misc import *
from .infer_gate_hist import get_circuit_gate_length
from ..platform.qcircuit_util import get_entanglement_bins
from ..platform.simulation.qcircuit_sim import schmidt_rank_vector
from .infer_compilation import generate_comp_tensors

from joblib import Parallel, delayed
import qiskit.quantum_info as qi

# %% ../../src/inference/infer_srv.ipynb 4
def get_all_srvs(num_of_qubits):
    srvs = [x for x in itertools.product(*([[1,2]]*num_of_qubits))]
    srvs = np.array(srvs, dtype=int)[np.sum(srvs, axis=1)!=num_of_qubits+1].tolist()
    srvs = sorted(srvs, key=lambda x: sum(x))
    return srvs

# %% ../../src/inference/infer_srv.ipynb 5
def generate_srv_tensors(pipeline, prompt, samples, system_size, num_of_qubits, max_gates, g, no_bar=True, unique=False, auto_batch_size=512):
    if samples==0:
        out_tensor = torch.zeros((0, system_size, max_gates))
        return out_tensor
    
    #----------------------
    #prepare condtions
    
    prompt = str(prompt)
    c = pipeline.text_encoder.tokenize_and_push_to_device(prompt)

    #----------------------
    #sample and post process to tensor encodings

    batch_samples = [auto_batch_size] * int(np.floor(samples/auto_batch_size))
    if samples % auto_batch_size > 0: batch_samples.append(samples % auto_batch_size)
    if len(batch_samples) == 0: batch_samples.append(samples)

    out_tensor_list = []
    for batch_sample in batch_samples:     
        
        c_batch = c.repeat(batch_sample, *[1]*(c.dim()-1))
        
        latents = torch.randn((c_batch.shape[0], pipeline.model.clr_dim, system_size, max_gates))    
        out_tensor = pipeline(latents=latents, c=c_batch, g=g, no_bar=no_bar, enable_guidance=True)   
        out_tensor_list.append(out_tensor)
        
    out_tensor = torch.cat(out_tensor_list)
    out_tensor = pipeline.model.invert_clr(out_tensor)
    out_tensor = out_tensor[:, :num_of_qubits]
    
    if unique: out_tensor = torch.unique(out_tensor, dim=0)
    
    if not no_bar: print(f"[INFO]: (generate_srv_tensors) Generated {'unique_cnt ' if unique else ''}{out_tensor.shape[0]} tensors")

    return out_tensor

# %% ../../src/inference/infer_srv.ipynb 7
def convert_tensors_to_srvs(out_tensor, gate_pool, sort_srv=False, place_barrier=False, n_jobs=1): 
    qc_list, error_cnt = convert_tensors_to_circuits(out_tensor, gate_pool=gate_pool, place_barrier=place_barrier)
    
    srv_list = []
    
    #---------------------------------------------
    # This is a bottle-neck for more qubits, speed up with async
    
    if n_jobs > 1:
        assert sort_srv == False
        
        f = lambda qc: schmidt_rank_vector(qi.DensityMatrix(qc))
        # srv_list = Parallel(n_jobs=n_jobs, prefer="threads")(delayed(f)(qc) for qc in qc_list)  #prefer="threads"
        srv_list = Parallel(n_jobs=n_jobs)(delayed(f)(qc) for qc in qc_list) 
    
    else:  
        for qc in qc_list:            
            srv = schmidt_rank_vector(qi.DensityMatrix(qc))
            
            if sort_srv: srv = sorted(srv)      
            srv_list.append(srv)  
  
    return qc_list, error_cnt, srv_list

# %% ../../src/inference/infer_srv.ipynb 9
def get_srv_accuracy(srv_list, target_srv):
    if not isinstance(srv_list  , (torch.Tensor, torch.IntTensor, torch.FloatTensor, torch.LongTensor)):   srv_list = torch.tensor(srv_list)
    if not isinstance(target_srv, (torch.Tensor, torch.IntTensor, torch.FloatTensor, torch.LongTensor)): target_srv = torch.tensor(target_srv, device=srv_list.device)
    
    srv_uniques, srv_uniques_cnt = torch.unique(srv_list, dim=0, return_counts=True)
    
    if srv_uniques.numel() == 0: return 0

    comp  = torch.all(target_srv==srv_uniques, dim=1)
    index = comp.nonzero().squeeze() 
        
    if index.dim() == 0: correct_srv_percentage = srv_uniques_cnt[index]/srv_uniques_cnt.sum()           
    else:                correct_srv_percentage = 0 
 
    return correct_srv_percentage

# %% ../../src/inference/infer_srv.ipynb 11
def true_sample_bin_dist(samples_per_bin, bin_size):
    true_samples     = [max(samples_per_bin//bin_size, 1) for i in range(bin_size)]

    if samples_per_bin-sum(true_samples) > 0:
        true_samples[0] += (samples_per_bin-sum(true_samples))
    
    # assert sum(true_samples)==samples_per_bin
    # assert len(true_samples)==bin_size

    # print(f"{true_samples=}")
    
    return true_samples

# %% ../../src/inference/infer_srv.ipynb 12
def test_srv_clr_distribution_bin_samples(pipeline, samples_per_bin, system_size, num_of_qubits, max_gates, g, gate_pool, silent=False, device="cpu", U=None, prompt_mod: callable=lambda c: c,
                                         only_diag=False, n_jobs=1):
    dist_srvs = get_all_srvs(num_of_qubits)  
    cond_srvs = dist_srvs

    values = torch.zeros((len(cond_srvs), len(dist_srvs)), device=device)

    #---------------------

    ent_bins, ent_labels = get_entanglement_bins(num_of_qubits)

    i = 0
    
    for ent_bin in tqdm(ent_bins, total=len(ent_bins)):

        true_samples = true_sample_bin_dist(samples_per_bin, len(ent_bin))
        
        for ind,srv in tqdm(enumerate(ent_bin), total=len(ent_bin)): 
            if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(srv), U, true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
            else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(srv),    true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)

            qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)
 
            if only_diag:
                values[i, i] = get_srv_accuracy(svr_list, srv)
            else:
                for j, dist_srv in enumerate(dist_srvs):      
                    values[i, j] = get_srv_accuracy(svr_list, dist_srv)
    
            if not silent:
                print(f"{cond_srv}:      unique_cnt {out_tensor.unique(dim=0).shape[0]}      error_cnt {error_cnt}      acc {values[i, i]:.2f}")

            i += 1
            
    return values

# %% ../../src/inference/infer_srv.ipynb 13
def test_srv_clr_distribution(pipeline, samples_per_srv, system_size, num_of_qubits, max_gates, g, gate_pool, silent=False, device="cpu", U=None, prompt_mod: callable=lambda c: c,
                             dist_srvs=None, cond_srvs=None, only_diag=False, n_jobs=1):
    if not exists(dist_srvs):
        dist_srvs = get_all_srvs(num_of_qubits)  
        
    if not exists(cond_srvs):
        cond_srvs = dist_srvs

    values = torch.zeros((len(cond_srvs), len(dist_srvs)), device=device)

    #---------------------
      
    for i, cond_srv in tqdm(enumerate(cond_srvs), total=len(cond_srvs)):    

        if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(cond_srv), U, samples_per_srv, system_size, num_of_qubits, max_gates, g=g, unique=False)
        else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(cond_srv),    samples_per_srv, system_size, num_of_qubits, max_gates, g=g, unique=False)
        
        qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)
                 
        if only_diag:
            values[i, i] = get_srv_accuracy(svr_list, srv)
        else:
            for j, dist_srv in enumerate(dist_srvs):      
                values[i, j] = get_srv_accuracy(svr_list, dist_srv)

        if not silent:
            print(f"{cond_srv}:      unique_cnt {out_tensor.unique(dim=0).shape[0]}      error_cnt {error_cnt}      acc {values[i, i]:.2f}")

    return values

# %% ../../src/inference/infer_srv.ipynb 14
def test_guidance_dep(pipeline, srvs, samples, system_size, num_of_qubits, max_gates, gs, gate_pool, prompt_mod: callable=lambda c: c, U=None, n_jobs=1):        
    guidance_dep_out = []

    for srv in srvs:
        unique_percentage_list      = []
        error_cnt_percentage_list   = []
        correct_srv_percentage_list = []
        
        for g in tqdm(gs):   
            if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(srv), U, samples, system_size, num_of_qubits, max_gates, g=g, unique=False)
            else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(srv),    samples, system_size, num_of_qubits, max_gates, g=g, unique=False)
                        
            #---------------------------------
            #calculate the copy percentage, dataset and sample? 
            
            unique_percentage  = out_tensor.unique(dim=0).shape[0]/out_tensor.shape[0]             
            unique_percentage_list.append(unique_percentage)
                            
            #---------------------------------
            #decode tensors, get srv
            
            qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)
            error_cnt_percentage_list.append(error_cnt/out_tensor.shape[0])
            
            #---------------------------------
            #record the correct number       
            
            correct_srv_percentage = get_srv_accuracy(svr_list, srv)   
            correct_srv_percentage_list.append(correct_srv_percentage)

        guidance_dep_out.append((unique_percentage_list, error_cnt_percentage_list, correct_srv_percentage_list))
    
    return guidance_dep_out

# %% ../../src/inference/infer_srv.ipynb 15
def test_srv_acc_vs_length(pipeline, samples, system_size, num_of_qubits, max_gates, g, gate_pool, prompt_mod: callable=lambda c: c, U=None, n_jobs=1):
    ent_bins, ent_labels = get_entanglement_bins(num_of_qubits)
    
    ent_ls   = []
    ent_accs = []
    ent_cnts = []
    
    for ent_bin in tqdm(ent_bins, total=len(ent_bins)):
        ls_acc = dict() #keep track over bins
        ls_cnt = dict()

        true_samples = true_sample_bin_dist(samples_per_bin, len(ent_bin))
        
        for ind,srv in enumerate(ent_bin): 
            if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(srv), U, true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
            else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(srv),    true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
                                   
            qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)     

            lengths = get_circuit_gate_length(qc_list) #work in qc space to check only non errors

            if lengths.numel() < 1: continue
            
            for l in lengths.unique(): #range(lengths.min(), lengths.max()):
                indices = (lengths==l).nonzero().squeeze() 
                               
                if indices.numel() > 0:                    
                    srvs = torch.tensor(svr_list)[indices] 
                    if indices.dim() == 0: srvs = srvs.unsqueeze(0)
                                            
                    acc = get_srv_accuracy(srvs, srv)
                    
                    #----------    
                    t = ls_acc.pop(l, [])
                    t.append(acc)
                    ls_acc[l] = t
                    
                    t = ls_cnt.pop(l, 0)
                    t += srvs.shape[0]
                    ls_cnt[l] = t
                           
        ls   = sorted(ls_acc)                   # sorted keys (l)
        accs = [np.mean(ls_acc[l]) for l in ls] # average acc per l
        cnts = [np.sum(ls_cnt[l]) for l in ls]
        
        ent_ls.append(ls)
        ent_accs.append(accs)
        ent_cnts.append(cnts)

    return ent_ls, ent_accs, ent_cnts, ent_labels

# %% ../../src/inference/infer_srv.ipynb 16
def test_srv_acc_vs_maxLength(pipeline, samples_per_bin, system_size, num_of_qubits, max_gates_list, g, gate_pool, prompt_mod: callable=lambda c: c, U=None, n_jobs=1):
    ent_bins, ent_labels = get_entanglement_bins(num_of_qubits)
    
    ent_accs = []
    for ent_bin in tqdm(ent_bins, total=len(ent_bins)):
        
        true_samples = true_sample_bin_dist(samples_per_bin, len(ent_bin))
        
        bin_accs = []
        for max_gates in max_gates_list:

            accs = []
            for ind,srv in enumerate(ent_bin): 
                if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(srv), U, true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
                else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(srv),    true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
                                   
                qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)     

                acc = get_srv_accuracy(svr_list, srv)
                
                accs.append(acc)       
            bin_accs.append(np.mean(accs))     
        ent_accs.append(bin_accs)
        
    return ent_accs, ent_labels

# %% ../../src/inference/infer_srv.ipynb 17
def test_srv_length_distribution(pipeline, samples_per_bin, system_size, num_of_qubits, max_gates, g, gate_pool, silent=False, U=None, prompt_mod: callable=lambda c: c, n_jobs=1):
    ent_bins, ent_labels = get_entanglement_bins(num_of_qubits)

    ls = []
    
    for ent_bin in tqdm(ent_bins, total=len(ent_bins)):

        true_samples = true_sample_bin_dist(samples_per_bin, len(ent_bin))
        
        bin_ls = []
        
        for ind,srv in tqdm(enumerate(ent_bin), total=len(ent_bin)): 
            if exists(U): out_tensor = generate_comp_tensors(pipeline, prompt_mod(srv), U, true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)
            else:         out_tensor = generate_srv_tensors( pipeline, prompt_mod(srv),    true_samples[ind], system_size, num_of_qubits, max_gates, g=g, unique=False)

            qc_list, error_cnt, svr_list = convert_tensors_to_srvs(out_tensor, gate_pool, n_jobs=n_jobs)

            qc_ls = get_circuit_gate_length(qc_list) #tensor [qcs]
            bin_ls.append(qc_ls)

        ls.append(torch.cat(bin_ls))

    return ls #[ent_bins, num_of_non_err_samples]

# %% ../../src/inference/infer_srv.ipynb 19
def plot_srv_clr_distribution_hist(values, samples, num_of_qubits, save=False, dist_srvs=None, cond_srvs=None):
    if not exists(dist_srvs):
        dist_srvs = get_all_srvs(num_of_qubits)  
        
    if not exists(cond_srvs):
        cond_srvs = dist_srvs

    n         = len(dist_srvs)
    values    = values.cpu()
    
    fig = plt.figure(figsize=(12,12))#, constrained_layout=True) 
    plt.title(f"Generated samples per condition: {samples}")
    plt.ylabel(r"Condition")
    plt.xlabel(r"Generated distribution")

    #--------------------------------------------
    if num_of_qubits < 6 or 0:
        plt.yticks(range(len(cond_srvs)), [str(b) for b in cond_srvs])
        plt.xticks(range(n), [str(b) for b in dist_srvs], rotation=90 if n>3 else 0)
    else:
        plt.yticks([])
        plt.xticks([])

    #--------------------------------------------
    plt.imshow(values, vmin=0, vmax=1)
    # plt.imshow(values.cpu(), norm="log")
    plt.colorbar()

    #--------------------------------------------
    #print acc
    x_shift = 1*40 if num_of_qubits==5 else 0
    if num_of_qubits < 4:
        for i in range(n): 
            plt.text(x_shift+i, i, f"{values[i, i]:0.2f}", color='black', ha='center', va='center', fontsize="large")

    #--------------------------------------------
    #draw rects
    off = 0.5
    for i in range(2, num_of_qubits):
        w = scipy.special.comb(num_of_qubits, i, exact=True)
        plt.gca().add_patch(plt.Rectangle((off, off), w, w, ls="-", ec="white", fc="none")) #, transform=plt.gca().transAxes))
        off += w

    #--------------------------------------------
    #print average acc for rects
    off = 0
    for i in [0]+list(range(2, num_of_qubits+1)):
        w = scipy.special.comb(num_of_qubits, i, exact=True)   
        d1 = off
        d2 = d1 + w
        mean_acc = values[d1:d2, d1:d2].diag().mean()
        plt.text(off+2*w/3, off+w/7, f"{mean_acc:0.2f}", color='red', ha='center', va='center', fontsize="x-large")   
        off += w

    #--------------------------------------------
    if save:
        plt.savefig('plot_srv_clr_distribution_hist.svg', bbox_inches='tight')
      
    plt.show()

# %% ../../src/inference/infer_srv.ipynb 20
def plot_srv_clr_distribution_bin_accuracy(values, samples, num_of_qubits, save=False, plot_percentages=False, trainSet_srv=None):
    values = values.cpu().diag()
    ent_bins, ent_labels = get_entanglement_bins(num_of_qubits)
    
    n = sum(len(srvs) for srvs in ent_bins)
    x = np.arange(n)  # the label locations
    width = 0.8

    #------------------------
    fig = plt.figure(figsize=(6.6, 4), constrained_layout=True) 
    # plt.title(f"Generated samples per condition: {samples}", fontsize=14)
    plt.ylabel(r"Accuracy", fontsize=25)
    plt.yticks(fontsize=14)
    plt.xticks([])
 
    i = 0
    for j,(label, srvs) in enumerate(zip(ent_labels, ent_bins)):   
        label = f"{sum(srvs[0])-num_of_qubits}"
        incre = len(srvs)
        rects = plt.bar(x[i:i+incre], values[i:i+incre], width, label=label)
        i += incre
        if plot_percentages: plt.gca().bar_label(rects, padding=3, fmt="%0.2f")

    ncols = len(ent_labels)//2+1 if len(ent_labels) > 5 else len(ent_labels)
    leg1 = plt.legend(loc="lower center", fontsize=14, ncols=ncols, title="# of entangled qubits:", title_fontsize=14,bbox_to_anchor=(0.5, 1.01))
    ax = fig.add_artist(leg1)
    
    if exists(trainSet_srv):
        if trainSet_srv.shape[-1]==num_of_qubits:
            srvs = []
            for s in ent_bins: srvs.extend(s)
            
            dataset_percentages = [get_srv_accuracy(trainSet_srv, srv).cpu() for srv in srvs]
            xmin = x - width*0.55
            xmax = x + width*0.55            
            ag = plt.hlines(dataset_percentages, xmin, xmax, label="Random sampling" , color="black", linestyle="-", linewidths=2.3)
        
            plt.legend(handles=[ag], fontsize=14, frameon=False)

    ymin, ymax = plt.ylim()    
    plt.ylim(ymin, ymax+0.04)

    if save:
        plt.savefig(f"plot_srv_clr_distribution_bin_accuracy.svg", bbox_inches='tight', transparent=True)
        
    plt.show()

# %% ../../src/inference/infer_srv.ipynb 21
def plot_guidance_dep(srvs, gs, guidance_dep_out, samples, save=False):    
    assert len(srvs) == len(guidance_dep_out)

    n = len(srvs)
    fig, axs = plt.subplots(1, n, figsize=(12, 5), squeeze=False, constrained_layout=True)   
    fig.suptitle(fr"Generated {samples} samples per $g$ and SRV")
    
    for i,srv in enumerate(srvs):  
        unique_percentage_list, error_cnt_percentage_list, correct_srv_percentage_list = guidance_dep_out[i]
                     
        #---------------------------------
        #plot now gs vs the numbers   
            
        plt.sca(axs[0, i])
        plt.xlabel(r"Guidance scale $g$")     
        plt.title(f"SRV = {srv}")
        plt.plot(gs, unique_percentage_list     , label="Unique tensors percentage")
        plt.plot(gs, error_cnt_percentage_list  , label="Error circuits percentage")
        plt.plot(gs, correct_srv_percentage_list, label="Correct SRV percentage")
        
        if i == (n-1): plt.legend() 

    if save:
        plt.savefig("plot_guidance_dep.svg", bbox_inches='tight', transparent=True)
    
    plt.show()

# %% ../../src/inference/infer_srv.ipynb 22
def plot_srv_acc_vs_length(ent_ls, ent_accs, ent_cnts, ent_labels, samples, plot_dist=True, save=False):
    fig, axs = plt.subplots(2 if plot_dist else 1, 1, figsize=(12, 7), squeeze=False, constrained_layout=True) 

    #-------------------
    plt.sca(axs[0,0])
    plt.title(f"Generated samples per entanglement: {samples}")
    plt.ylabel("Accuracy")
    plt.xlabel("Gate number")
    for i,ent_label in enumerate(ent_labels): 
        plt.plot(ent_ls[i], ent_accs[i], label=f"{ent_label}")
    plt.legend()

    #-------------------
    if plot_dist:
        plt.sca(axs[1,0])
        plt.title(f"Used samples per l to calculate accuracy, should match gate distribution")
        plt.ylabel("Used samples")
        plt.xlabel("Gate number")
        for i,ent_label in enumerate(ent_labels): 
            plt.plot(ent_ls[i], ent_cnts[i], label=f"{ent_label}")
        plt.legend()

    #-------------------
    if save:
        plt.savefig('plot_srv_acc_vs_length.svg', bbox_inches='tight')
         
    plt.show()

# %% ../../src/inference/infer_srv.ipynb 23
def plot_srv_acc_vs_maxLength(ent_accs, ent_labels, max_gates_list, samples, plot_dist=True, save=False):
    fig = plt.figure(figsize=(12, 4), constrained_layout=True) 
    
    plt.title(f"Generated samples per maxGates per entanglement: {samples}")
    plt.ylabel("Accuracy")
    plt.xlabel("Max number of gates / tensor size")
    plt.xticks(max_gates_list)
    
    for ent_acc,ent_label in zip(ent_accs, ent_labels): 
        plt.plot(max_gates_list, ent_acc, label=f"{ent_label}")
        
    plt.legend()

    if save:
        plt.savefig('plot_srv_acc_vs_length.svg', bbox_inches='tight')
         
    plt.show()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/config_loader.ipynb.

# %% auto 0
__all__ = ['class_to_str', 'load_config', 'config_to_dict', 'save_dataclass_yaml', 'save_dict_yaml', 'get_obj_from_str',
           'instantiate_from_config', 'load_model_from_config']

# %% ../src/config_loader.ipynb 3
from .imports import *
from omegaconf import OmegaConf

# %% ../src/config_loader.ipynb 5
def class_to_str(cls):
    return str(cls)[8:-2]

# %% ../src/config_loader.ipynb 6
def load_config(file_path):
    return OmegaConf.load(f"{file_path}")

# %% ../src/config_loader.ipynb 7
def config_to_dict(config):
    return OmegaConf.to_container(config)

# %% ../src/config_loader.ipynb 8
def save_dataclass_yaml(data_obj, file_path):
    conf = OmegaConf.structured(data_obj)
    with open(file_path, 'w') as f:
        OmegaConf.save(config=conf, f=f)

# %% ../src/config_loader.ipynb 9
def save_dict_yaml(dict_obj, file_path):
    conf = OmegaConf.create(dict_obj)
    with open(file_path, 'w') as f:
        OmegaConf.save(config=conf, f=f)

# %% ../src/config_loader.ipynb 14
def get_obj_from_str(string, reload=False):
    module, cls = string.rsplit(".", 1)
    if reload:
        module_imp = importlib.import_module(module)
        importlib.reload(module_imp)
    return getattr(importlib.import_module(module, package=None), cls)

# %% ../src/config_loader.ipynb 15
def instantiate_from_config(config):
    if not "target" in config: raise KeyError("Expected key `target` to instantiate.")
    if not "params" in config: print(f"[WARNING] Expected key `params` to instantiate.")
    return get_obj_from_str(config["target"])(**config.get("params", dict()))

# %% ../src/config_loader.ipynb 16
def load_model_from_config(config, ckpt, device):
    
    print(f"Loading model from {ckpt}")
    pl_sd = torch.load(ckpt, map_location=torch.device(device).type, weights_only=True)
          
    model = instantiate_from_config(config.model)
    
    sd = pl_sd["state_dict"]
    m, u = model.load_state_dict(sd, strict=True)
    
    return model.to(device)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/models/layers.ipynb.

# %% auto 0
__all__ = ['DownBlock2D', 'UpBlock2D', 'ResDownBlock2D', 'ResUpBlock2D', 'ResBlock2D', 'ResBlock2D_Conditional', 'FeedForward',
           'PositionalEncoding', 'TimeEmbedding', 'PositionalEncodingTransposed', 'PositionalEncoding2D',
           'PositionalEncoding2DSpaceOnly']

# %% ../../src/models/layers.ipynb 3
from ..imports import *

# %% ../../src/models/layers.ipynb 5
class DownBlock2D(nn.Module):  
    """A 2d down scale block."""
    def __init__(self, in_ch, out_ch, kernel_size=2, stride=2, padding=0, use_conv=True):
        super().__init__()  
        self.use_conv = use_conv                
        if self.use_conv: 
            self.conv1  = nn.Conv2d(in_ch, out_ch, kernel_size=kernel_size, stride=stride, padding=padding)           
        else:       
            self.avg_pool = nn.AvgPool2d(kernel_size=kernel_size, stride=stride, padding=padding)    
            self.convId   = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1, padding="same") if in_ch!=out_ch else nn.Identity()   
        
    def forward(self, x):               
        if self.use_conv:
            #x = F.pad(x, pad=(0,0,0,1), mode="constant", value=0) #for 2d: pad=(0,1,0,1)
            x = self.conv1(x)          
        else:
            x = self.avg_pool(x)
            x = self.convId(x)           
        return x

# %% ../../src/models/layers.ipynb 6
class UpBlock2D(nn.Module):  
    """A 2d up scale block."""
    def __init__(self, in_ch, out_ch, kernel_size=2, stride=2, padding=0, use_conv=True):
        super().__init__()  
        self.use_conv  = use_conv                 
        self.up_sample = nn.Upsample(scale_factor=kernel_size)           
        if self.use_conv: 
            self.conv1  = nn.Conv2d(in_ch, out_ch, kernel_size=(1,3), stride=1, padding="same")  
        else:             
            self.convId = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1, padding="same") if in_ch!=out_ch else nn.Identity()   
            
    def forward(self, x):        
        x = self.up_sample(x)       
        if self.use_conv: 
            x = self.conv1(x)          
        else:             
            x = self.convId(x)                
        return x

# %% ../../src/models/layers.ipynb 7
class ResDownBlock2D(nn.Module):  
    """A 2d residual down scale block."""
    def __init__(self, in_ch, out_ch, kernel_size=2, stride=2, padding=0):
        super().__init__()     
        self.act    = nn.SiLU() 
        self.conv1  = nn.Conv2d( in_ch, out_ch, kernel_size=kernel_size, stride=stride, padding=padding) 
        self.norm   = torch.nn.GroupNorm(num_groups=16, num_channels=out_ch, eps=1e-5, affine=True)        
        self.convId = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1, padding="same") if in_ch!=out_ch else lambda x:x
        self.down   = nn.AvgPool2d(kernel_size=kernel_size, stride=stride, padding=padding)
               
    def forward(self, x):         
        r1 = self.conv1(x)
        r1 = self.norm(r1)
        r1 = self.act(r1)
        
        r2 = self.convId(x)
        r2 = self.down(r2)
        return self.act(r1 + r2)

# %% ../../src/models/layers.ipynb 8
class ResUpBlock2D(nn.Module):  
    """A 2d residual up scale block."""
    def __init__(self, in_ch, out_ch, kernel_size=2, stride=2, padding=0):
        super().__init__()     
        self.act    = nn.SiLU()
        self.conv1  = nn.ConvTranspose2d(in_ch, out_ch, kernel_size=kernel_size, stride=stride, padding=padding) 
        self.norm   = torch.nn.GroupNorm(num_groups=16, num_channels=out_ch, eps=1e-6, affine=True)        
        self.convId = nn.Conv2d(in_ch, out_ch, kernel_size=1, stride=1, padding="same") if in_ch!=out_ch else nn.Identity()
        self.up     = nn.Upsample(scale_factor=kernel_size)
               
    def forward(self, x):         
        r1 = self.conv1(x)
        r1 = self.norm(r1)
        r1 = self.act(r1)
        
        r2 = self.convId(x)
        r2 = self.up(r2)
        return self.act(r1 + r2)

# %% ../../src/models/layers.ipynb 10
class ResBlock2D(nn.Module):
    """A 2d residual block."""
    def __init__(self, in_ch, out_ch, kernel_size, skip=True):
        super().__init__()             
        self.act   = nn.SiLU()             
        self.conv1 = nn.Conv2d( in_ch, out_ch, kernel_size=kernel_size, stride=1, padding ="same")        
        self.conv2 = nn.Conv2d(out_ch, out_ch, kernel_size=kernel_size, stride=1, padding ="same")
        
        self.norm1  = torch.nn.GroupNorm(num_groups=32, num_channels=in_ch) #, eps=1e-6, affine=True)         
        self.norm2  = torch.nn.GroupNorm(num_groups=32, num_channels=out_ch) #, eps=1e-6, affine=True)  
        
        self.skip  = skip 
        if self.skip:        
            self.skip_connection= nn.Conv2d(in_ch,  out_ch, kernel_size=1, stride=1, padding ="same") if in_ch!=out_ch else nn.Identity()
         
    def forward(self, x):      
        
        #in layers
        h = self.norm1(x) 
        h = self.act(h)
        h = self.conv1(h)  
        
        #out layers
        h = self.norm2(h) 
        h = self.act(h)
        h = self.conv2(h)
    
        #----
    
        if not self.skip: return h  
                   
        return self.skip_connection(x) + h

# %% ../../src/models/layers.ipynb 11
class ResBlock2D_Conditional(nn.Module):
    """A 2d residual block with input of a time-step $t$ embedding."""
    def __init__(self, in_ch, out_ch, t_emb_size, kernel_size, skip=True):
        super().__init__()             
        self.act   = nn.SiLU()             
        self.conv1 = nn.Conv2d( in_ch, out_ch, kernel_size=kernel_size, stride=1, padding ="same")        
        self.conv2 = nn.Conv2d(out_ch, out_ch, kernel_size=kernel_size, stride=1, padding ="same")
        
        self.norm1  = torch.nn.GroupNorm(num_groups=32, num_channels=in_ch) #, eps=1e-6, affine=True)         
        self.norm2  = torch.nn.GroupNorm(num_groups=32, num_channels=out_ch) #, eps=1e-6, affine=True)  
        
        self.skip  = skip 
        if self.skip:        
            self.skip_connection= nn.Conv2d(in_ch,  out_ch, kernel_size=1, stride=1, padding ="same") if in_ch!=out_ch else nn.Identity()
        
        self.t_proj = nn.Linear(t_emb_size, out_ch)    
               
    def forward(self, x, t_emb):      
        
        #in layers
        h = self.norm1(x) 
        h = self.act(h)
        h = self.conv1(h)  
        
        #embed
        h = h + self.t_proj(t_emb)[:, :, None, None]
        
        #out layers
        h = self.norm2(h) 
        h = self.act(h)
        h = self.conv2(h)
    
        #----
    
        if not self.skip: return h  
                   
        return self.skip_connection(x) + h

# %% ../../src/models/layers.ipynb 13
class FeedForward(nn.Module):
    """A small dense feed-forward network as used in `transformers`."""
    def __init__(self, in_ch, out_ch, inner_mult=1):
        super().__init__()
        self.proj1 = nn.Linear(in_ch, in_ch*inner_mult)   
        self.proj2 = nn.Linear(in_ch*inner_mult, out_ch) 
        self.act   = nn.SiLU()
        
    def forward(self, x):
        x = self.proj1(x)
        x = self.act(x)
        x = self.proj2(x) 
        return x     

# %% ../../src/models/layers.ipynb 16
class PositionalEncoding(nn.Module):
    """An absolute pos encoding layer."""
    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, d_model)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len , embedding_dim]``
        """
        x = x + self.pe[None, :x.size(1)]
        return self.dropout(x)

# %% ../../src/models/layers.ipynb 17
class TimeEmbedding(PositionalEncoding):
    """A time embedding layer"""
    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 5000):
        super().__init__(d_model, dropout, max_len)             
        self.ff = FeedForward(d_model, d_model)  
       
    def forward(self, t: torch.Tensor):       
        x = self.pe[t]       
        x = self.ff(x)               
        return self.dropout(x)

# %% ../../src/models/layers.ipynb 18
class PositionalEncodingTransposed(PositionalEncoding):
    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 5000):
        super().__init__(d_model, dropout, max_len)        
        self.pe = torch.permute(self.pe, (1, 0)) # [max_len, d_model] to [d_model, max_len]
               
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[batch_size, embedding_dim, seq_len]``
        """
        x = x + self.pe[None, :, :x.size(2)]
        return self.dropout(x)

# %% ../../src/models/layers.ipynb 19
class PositionalEncoding2D(PositionalEncodingTransposed):
    """A 2D absolute pos encoding layer."""
    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 5000):
        super().__init__(d_model=d_model//2, dropout=dropout, max_len=max_len)    
        self.d_model_half = d_model//2    
        # self.proj         = nn.Conv2d(d_model, d_model, kernel_size=1, stride=1, padding ="same")      
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[batch_size, gate_color, space , time]``
        """
                                          
        p1 = self.pe[None, :, :x.size(2),       None] #space encoding
        p2 = self.pe[None, :,       None, :x.size(3)] #time encoding 
                                
        x[:, :self.d_model_half] = x[:, :self.d_model_half] + p1
        x[:, self.d_model_half:] = x[:, self.d_model_half:] + p2
        
        # x = self.proj(x)
        
        return self.dropout(x)

# %% ../../src/models/layers.ipynb 20
class PositionalEncoding2DSpaceOnly(PositionalEncodingTransposed):
    def __init__(self, d_model: int, dropout: float = 0.0, max_len: int = 5000):
        super().__init__(d_model=d_model, dropout=dropout, max_len=max_len)      
    
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[batch_size, gate_color, space , time]``
        """
                                          
        p1 = self.pe[None, :, :x.size(2), None] #space encoding        
        return self.dropout(x + p1)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../src/pipeline/diffusion_pipeline.ipynb.

# %% auto 0
__all__ = ['DiffusionPipeline']

# %% ../../src/pipeline/diffusion_pipeline.ipynb 2
from ..imports import *
from ..scheduler.scheduler import Scheduler
from .pipeline import Pipeline
from ..config_loader import *
from ..models.config_model import Config_Model

from huggingface_hub import snapshot_download

# %% ../../src/pipeline/diffusion_pipeline.ipynb 3
class DiffusionPipeline(Pipeline):   
    """A `Pipeline` for diffusion models. Implements train and inference functions. Diffusion parameters are defined inside a `Scheduler` object."""
    non_blocking = False
    
    def __init__(self, 
                 scheduler: Scheduler,   
                 model: nn.Module,
                 text_encoder: nn.Module,
                 device: torch.device,
                 enable_guidance_train = True,
                 guidance_train_p = 0.1,
                 cached_text_enc = True
                ):    
        super().__init__(model, device)
        self.scheduler = scheduler
        self.scheduler.to_device(device)    
        
        self.text_encoder = text_encoder
        self.text_encoder.eval()
        
        self.enable_guidance_train = enable_guidance_train           
        self.guidance_train_p      = guidance_train_p
           
        self.cached_text_enc = cached_text_enc        
        self.empty_token     = self.text_encoder.empty_token
        
        if cached_text_enc:                       
            def cached_empty_token_fn(c):
                if   c.dim() == 1: return self.text_encoder.cached_empty_token_index  # yields then a list of ints       
                elif c.dim() == 2: return self.empty_token.expand(c.shape)            # tokenized input      
                else: raise NotImplementedError("")
            
            self.empty_token_fn = cached_empty_token_fn    
                        
        else:
            self.empty_token_fn = lambda c: self.empty_token.expand(c.shape) # for own clip  

    #------------------------------------
         
    add_config = {}
        
    def params_config(self, save_path: str):         
        params_config = {}
                
        params_config["scheduler"]    = self.scheduler.get_config()
        params_config["model"]        = self.model.get_config(save_path=save_path+"model.pt")
        params_config["text_encoder"] = self.text_encoder.get_config(save_path=save_path+"text_encoder.pt")
        
        params_config["device"]                = str(self.device)
        params_config["enable_guidance_train"] = self.enable_guidance_train
        params_config["guidance_train_p"]      = self.guidance_train_p
        params_config["cached_text_enc"]       = self.cached_text_enc
        params_config["add_config"]            = self.add_config
        
        return params_config
    
    def store_pipeline(self, config_path: str, save_path: str):
        super().store_pipeline(config_path, save_path)
        config = self.get_config(save_path)
        save_dict_yaml(config, config_path+"config.yaml")
               
        #only store weights of these submodels
        self.model.store_model(config_path=None, save_path=save_path+"model.pt")
        self.text_encoder.store_model(config_path=None, save_path=save_path+"text_encoder.pt")
    
    @staticmethod
    def from_config_file(config_path, device: torch.device):    
        config = load_config(config_path+"config.yaml")   
        config = config_to_dict(config)

        if exists(device):
            config["params"]["device"]                        = device
            config["params"]["scheduler"]["params"]["device"] = device
        
        config["params"]["scheduler"] = instantiate_from_config(config["params"]["scheduler"])
 
        model_path                = config_path+"model.pt" if config["params"]["model"]["save_path"] is None else config["params"]["model"]["save_path"]
        config["params"]["model"] = Config_Model.from_config(config["params"]["model"], device, model_path)

        config["params"]["text_encoder"] = Config_Model.from_config(config["params"]["text_encoder"], device, config["params"]["text_encoder"]["save_path"])   
        add_config = config["params"].pop("add_config", None)

        pipeline = instantiate_from_config(config)
        
        if exists(pipeline.add_config):
            pipeline.gate_pool  = [gate for gate in add_config["dataset"]["params"]["gate_pool"]] 
            pipeline.add_config = add_config
        
        return pipeline


    @classmethod
    def from_pretrained(cls, repo_id: str, device: torch.device, **kwargs):  
        """Load a model pipeline directly from Huggingface."""
        model_path = snapshot_download(repo_id=repo_id, repo_type="model", allow_patterns=["*.pt", "*.yaml", "*.safetensors"], **kwargs) 
        pipeline   = cls.from_config_file(model_path+"/", device)  
        return pipeline
    
    #------------------------------------
    # Inference functions
    
    @torch.no_grad()
    def __call__(self, latents=None, c=None, seed=None, timesteps=None, no_bar=False, enable_guidance=True, g=7.5):        
        if exists(seed):      torch.manual_seed(seed)
        if exists(timesteps): self.scheduler.set_timesteps(self.timesteps)
            
        latents = latents.to(self.device)                  
        x0      = self.denoising(latents, c=c, no_bar=no_bar, enable_guidance=enable_guidance, g=g)  
        
        return x0
    
    @torch.no_grad()
    def latent_filling(self, org_latents: torch.Tensor, mask: torch.Tensor, c=None, enable_guidance=True, g=7.5, 
                       t_start_index=0, no_bar=False, return_predicted_x0=False, **kwargs):
        """mask: area with ones is going to be filled"""
        if   mask.dim() == 4: assert list(org_latents.shape) == list(mask.shape)     # diff mask per sample and channel
        elif mask.dim() == 3: assert list(org_latents.shape[1:]) == list(mask.shape) # diff mask per channel
        elif mask.dim() == 2: assert list(org_latents.shape[2:]) == list(mask.shape) # all same mask
        else:                 raise RuntimeError("")
    
        self.model.eval()
        self.text_encoder.eval()       
        self.scheduler.to_device(self.device)
                 
        c_emb = self.prepare_c_emb(c, enable_guidance, **kwargs)
    
        org_latents = org_latents.to(self.device, non_blocking=self.non_blocking)

        #---
        # ch to prep latent, init_image
        
        noise   = torch.randn(org_latents.shape, device=self.device)   
        
        if t_start_index > 0: latents = self.scheduler.add_noise(org_latents, noise, self.scheduler.timesteps[t_start_index])
        else:                 latents = noise
        
        #---

        timesteps = self.scheduler.timesteps[t_start_index:]
        
        if return_predicted_x0: predicted_x0 = list()
     
        for i, t in enumerate(tqdm(timesteps, disable=no_bar)):
            timesteps = (torch.ones((1)) * t).type(torch.int64).to(self.device, non_blocking=self.non_blocking)        

            #----------------------------------
            #gen the fill in part
            latents, x0 = self.denoising_step(latents, timesteps, c_emb=c_emb, enable_guidance=enable_guidance, g=g)
            
            #----------------------------------
            #create the part we don't fill 

            noisy_org_latents = org_latents
            
            if i < len(timesteps) - 1:            #else we are finished with denoising
                noise_timestep = timesteps[i + 1]
                # noise = torch.randn(org_latents.shape, device=self.device)     
                noisy_org_latents = self.scheduler.add_noise(noisy_org_latents, noise, noise_timestep) 
                
            #combine
            latents = (1.0-mask) * noisy_org_latents + mask * latents  

            #----------------------------------   
            if return_predicted_x0: 
                x0      = (1.0-mask) *  org_latents + mask * fill_x0  
                predicted_x0.append(x0)
        
        if return_predicted_x0: return latents.cpu(), predicted_x0     
        return latents.cpu()
    
    #------------------------------------
    # Helper functions
    
    def get_guidance_condition(self, c, enable_guidance):
        if not exists(c): return c      
        c = c.to(self.device)                
        if enable_guidance:             
            u = self.empty_token_fn(c).to(self.device)        
            c = torch.cat([u, c])            
        c = c.type(torch.int64) 
        return c

    def prepare_c_emb(self, c, enable_guidance, **kwargs):
        c     = self.get_guidance_condition(c, enable_guidance) 
        c_emb = self.text_encoder(c, pool=False)
        return c_emb
        
    @torch.no_grad()
    def denoising(self, latents: torch.Tensor, c=None, enable_guidance=True, g=7.5, t_start_index=0, no_bar=False, return_predicted_x0=False, **kwargs):
        self.model.eval()
        self.text_encoder.eval()
        self.scheduler.to_device(self.device)
    
        c_emb = self.prepare_c_emb(c, enable_guidance, **kwargs)
        
        latents = latents.to(self.device, non_blocking=self.non_blocking)
        
        if return_predicted_x0: predicted_x0 = list()
        
        for i, t in enumerate(tqdm(self.scheduler.timesteps[t_start_index:], disable=no_bar)):
            timesteps = (torch.ones((1)) * t).type(torch.int64).to(self.device, non_blocking=self.non_blocking)
                   
            latents, x0 = self.denoising_step(latents, timesteps, c_emb=c_emb, enable_guidance=enable_guidance, g=g, **kwargs)
        
            if return_predicted_x0: predicted_x0.append(x0.cpu())
        
        if return_predicted_x0: return latents.cpu(), predicted_x0     
        return latents.cpu()
  
    # @torch.no_grad()
    def denoising_step(self, latents: torch.Tensor, ts: Union[int, torch.IntTensor], c_emb: torch.Tensor=None, enable_guidance=True, g=7.5, **kwargs):    
        if enable_guidance:
            x = torch.cat([latents] * 2)     #uses batch layer combine here
            
            if ts.numel() > 1: chunk_ts = torch.cat([ts] * 2)
            else:              chunk_ts = ts
                
            eps_u, eps_c = self.model(x, chunk_ts, c_emb).chunk(2) 
            
            eps = self.CFG(eps_u, eps_c, g)
                    
        else:
            eps = self.model(latents, ts, c_emb)  
                 
        x = self.scheduler.step(eps, ts, latents)      
        return x.prev_sample, x.pred_original_sample
     
    guidance_sample_mode = "rescaled" # one of: normal, fastai, rescaled
        
    def CFG(self, eps_u, eps_c, g):
        """Apply Classifier-free-guidance sampling"""
        dim = list(range(1, eps_u.dim())) # reduce all but batches

        if self.guidance_sample_mode == "normal":   # from https://arxiv.org/pdf/2207.12598.pdf, w=g+1    
            eps = eps_u + g * (eps_c-eps_u)               

        elif self.guidance_sample_mode == "fastai":  # from fastAi less 11
            eps = eps_u + g*(eps_c-eps_u) * torch.linalg.vector_norm(eps_u, dim=dim, keepdim=True) / torch.linalg.vector_norm(eps_c-eps_u, dim=dim, keepdim=True)    
            eps = eps * torch.linalg.vector_norm(eps_u, dim=dim, keepdim=True) / torch.linalg.vector_norm(eps, dim=dim, keepdim=True)

        elif self.guidance_sample_mode == "rescaled": # from https://arxiv.org/pdf/2305.08891.pd
            phi = 0.7

            eps_cfg    = eps_u + g * (eps_c-eps_u)              
            eps_scaled = eps_cfg * eps_c.std(dim=dim, keepdim=True) / eps_cfg.std(dim=dim, keepdim=True)
            eps        = phi * eps_scaled + (1.0-phi) * eps_cfg

        else: raise NotImplementedError("Use as guidance_sample_mode one of: 'normal', 'fastai' or 'rescaled'")

        return eps
        
    #------------------------------------
    # Training functions

    def train_on_epoch(self, data_loader: DataLoader, train=True):   
        self.scheduler.to_device(self.device, non_blocking=self.non_blocking)    
        super().train_on_epoch(data_loader, train)

    #@torch.autocast(device_type=device.type)
    def train_step(self, data, **kwargs): 
        latents, y = data                
        b, s, t = latents.shape          
        
        #start async memcpy
        latents = latents.to(self.device, non_blocking=self.non_blocking)  
        latents = self.model.embedd_clrs(latents)                   #this is only new tensor
               
        #do the cond embedding with CLIP                     
        y = y.to(self.device, non_blocking=self.non_blocking)  
        
        if self.enable_guidance_train: 
            rnd = torch.rand((b,), device=self.device)            
            rnd = (rnd > self.guidance_train_p).type(torch.int64)          # todo: change to bernoulli dist fn
            rnd = self.scheduler.unsqueeze_vector_to_shape(rnd, y.shape)   # e.g. [b, 1, 1]            
            y   = y * rnd + (1-rnd) * self.empty_token_fn(y)
               
        y_emb = self.text_encoder(y, pool=False)
                          
        #sample timesteps
        timesteps = torch.randint(low=0, high=self.scheduler.num_train_timesteps, size=(b,), device=self.device, dtype=torch.int64)

        #forward noising    
        noise = torch.randn(latents.shape, device=self.device)     
        noisy_latents = self.scheduler.add_noise(latents, noise, timesteps)

        #predict eps
        eps = self.model(noisy_latents, timesteps, y_emb)
            
        #comp mse
        loss = self.loss_fn(eps, noise)
        
        #log the loss
        return loss



# AUTOGENERATED! DO NOT EDIT! File to edit: ../src/metrics.ipynb.

# %% auto 0
__all__ = ['Metric', 'Mean', 'Accuracy']

# %% ../src/metrics.ipynb 3
from .imports import *
from .util import virtual

# %% ../src/metrics.ipynb 4
class Metric:
    """Base metric class."""
    def __init__(self, name: str, device): 
        self.name   = name
        self.device = torch.device(device)
        self.reset_state() 
    def __repr__(self): return f"{self.name}={self.result()}"

    def update_state(self, inp, tar=None): self.empty=False
    def reset_state(self): self.empty=True        
    
    @virtual
    def _eval(self, inp, tar): pass
    @virtual
    def result(self): pass 

# %% ../src/metrics.ipynb 5
class Mean(Metric):
    """Mean metric, used for loss .."""
    def __init__(self, name: str, device): super().__init__(name, device)   
    @torch.inference_mode()
    def update_state(self, inp: torch.Tensor, tar: torch.Tensor=None, weight: float=1):
        super().update_state(inp, tar)         
        val = self._eval(inp, tar)        
        self.weighted_sum += torch.sum(val * weight)
        self.weight       += weight * torch.numel(val)   
    @torch.inference_mode()
    def reset_state(self): 
        super().reset_state()
        self.weighted_sum = torch.tensor(0.0, device=self.device)
        self.weight       = torch.tensor(0.0, device=self.device)   
    def _eval(self, inp, tar): return inp  
    @torch.inference_mode()
    def result(self): return (self.weighted_sum/self.weight).cpu()

# %% ../src/metrics.ipynb 6
class Accuracy(Mean):   
    """Accuracy metric."""
    @torch.inference_mode()
    def _eval(self, inp, tar): return (inp==tar).float().mean()
